import os
import sys


class mute_stdout_stderr(object):
    """Context manager to redirect stdout and stderr to devnull. Only used to prevent terminal flooding in unittests."""
    def __enter__(self):
        self.outnull = open(os.devnull, 'w')
        self.old_stdout = sys.stdout
        self.old_stderr = sys.stderr
        sys.stdout = self.outnull
        sys.stderr = self.outnull
        return self

    def __exit__(self):
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr
        self.outnull.close()



from packaging import version
import pandas as pd
import numpy as np
import numpy.typing as npt
import logging
import scipy.ndimage as ndimage
import pathlib
from typing import Optional
from pytom_tm.tmjob import TMJob
from pytom_tm.mask import spherical_mask
from pytom_tm.angles import load_angle_list, convert_euler
from pytom_tm.io import read_mrc
from scipy.special import erfcinv
from scipy.optimize import curve_fit
from tqdm import tqdm


plotting_available = False
try:
    import matplotlib.pyplot as plt
    import seaborn as sns
    sns.set(context='talk', style='ticks')
    plotting_available = True
except ModuleNotFoundError:
    pass


def predict_tophat_mask(
        score_volume: npt.NDArray[float],
        output_path: Optional[pathlib.Path] = None,
        n_false_positives: int = 1,
        create_plot: bool = True
) -> npt.NDArray[bool]:
    """This function gets as input a score map and returns a peak mask as determined with a tophat transform.

    It does the following things:
     - calculate a tophat transform using scipy.ndimage.white_tophat() and a kernel
     ndimage.generate_binary_structure(rank=3, connectivity=1).
     - calculate a histogram of the transformed score map and take its log to focus more on small values
     - take second derivative of log(histogram) to find the region for fitting a Gaussian, where the second derivative
     switches from negative to positive the background noise likely breaks
     - use formula from excellent work of Rickgauer et al. (2017, eLife) which uses the error function to find the
     likelihood of false positives on the background Gaussian distribution:
            N**(-1) = erfc( theta / ( sigma * sqrt(2) ) ) / 2

    Parameters
    ----------
    score_volume: npt.NDArray[float]
        template matching score map
    output_path: Optional[pathlib.Path], default None
        if provided (and plotting is available), write a figure of the fit to the output folder
    n_false_positives: int, default 1
        number of false positive for error function cutoff calculation
    create_plot: bool, default True
        whether to plot the gaussian fit and cut-off estimation

    Returns
    -------
    peak_mask: npt.NDArray[bool]
        boolean mask with tophat filtered peak locations
    """
    tophat = ndimage.white_tophat(
        score_volume,
        structure=ndimage.generate_binary_structure(
            rank=3,
            connectivity=1
        )
    )
    y, bins = np.histogram(tophat.flatten(), bins=50)
    bin_centers = (bins[:-1] + bins[1:]) / 2
    x_raw, y_raw = bin_centers[2:], y[2:]  # discard first two points because of over-representation of zeros

    # take second derivative and discard inaccurate boundary value (hence the [2:])
    with np.errstate(divide='ignore'):
        y_log = np.log(y_raw)
        y_log[np.isinf(y_log)] = 0
    second_derivative = np.gradient(np.gradient(y_log))[2:]
    m1 = second_derivative[:-1] < 0  # where the derivative is negative
    sign = np.sign(second_derivative[1:] * second_derivative[:-1])
    # if there was a 0 in second derivative, there are now two 0s in sign
    # replace the 0,0 with the next value and previous value instead
    sign = np.where(sign == 0, np.roll(sign, -1), sign)
    sign = np.where(sign == 0, np.roll(sign, 1), sign)
    m2 = sign == -1  # switches from neg. to pos. and vice versa
    idx = (
            int(np.argmax(m1 & m2))  # first switch from negative to positive
            + 2  # +2 for 2nd derivative discarded part
            + 1  # +1 to include last value
    )
    x_fit, y_fit = x_raw[:idx], y_raw[:idx]

    def gauss(x, amp, mu, sigma):  # gaussian for fitting
        return amp * np.exp(-(x - mu) ** 2 / (2 * sigma ** 2))

    def log_gauss(x, amp, mu, sigma):  # log of gaussian for fitting
        return np.log(gauss(x, amp, mu, sigma))

    guess = np.array([y.max(), 0, score_volume.std()])  # rough initial guess for amp, mu and sigma
    coeff = curve_fit(gauss, x_fit, y_fit, p0=guess)[0]  # first fit regular gauss to get better guesses
    coeff_log = curve_fit(log_gauss, x_fit, np.log(y_fit), p0=coeff)[0]  # now go for accurate fit to log of gauss
    search_space = coeff_log[0] / (coeff_log[2] * np.sqrt(2 * np.pi))
    # formula Rickgauer et al. (2017, eLife): N**(-1) = erfc( theta / ( sigma * sqrt(2) ) ) / 2
    # we need to find theta (i.e. the cut-off)
    cut_off = erfcinv((2 * n_false_positives) / search_space) * np.sqrt(2) * coeff_log[2] + coeff_log[1]

    if plotting_available and output_path is not None and create_plot:
        fig, ax = plt.subplots()
        ax.scatter(x_raw, y_raw, label='tophat', marker='o')
        ax.plot(x_raw, gauss(x_raw, *coeff_log), label='pred', color='tab:orange')
        ax.axvline(cut_off, color='gray', linestyle='dashed', label='cut-off')
        ax.axvspan(x_fit[0], x_fit[-1], alpha=0.25, color='gray', label='fitted data')
        ax.set_yscale('log')
        ax.set_ylim(bottom=.1)
        ax.set_ylabel('Occurence')
        ax.set_xlabel('Tophat scores')
        ax.legend()
        plt.tight_layout()
        plt.savefig(output_path, dpi=600, transparent=False, bbox_inches='tight')

    peak_mask = (tophat > cut_off)

    return peak_mask


def extract_particles(
        job: TMJob,
        particle_radius_px: int,
        n_particles: int,
        cut_off: Optional[float] = None,
        n_false_positives: int = 1,
        tomogram_mask_path: Optional[pathlib.Path] = None,
        tophat_filter: bool = False,
        create_plot: bool = True
) -> tuple[pd.DataFrame, list[float, ...]]:
    """
    Parameters
    ----------
    job: pytom_tm.tmjob.TMJob
        template matching job to annotate particles from
    particle_radius_px: int
        particle radius to remove peaks with after a score has been annotated
    n_particles: int
        maximum number of particles to extract
    cut_off: Optional[float]
        manually override the automated score cut-off estimation, value between 0 and 1
    n_false_positives: int
        tune the number of false positives to be included for automated error function cut-off estimation:
        should be an integer >= 1
    tomogram_mask_path: Optional[pathlib.Path]
        path to a tomographic binary mask for extraction
    tophat_filter: bool
        attempt to only select sharp peaks with the tophat filter
    create_plot: bool, default True
        flag for creating extraction plots

    Returns
    -------
    dataframe, scores: tuple[pd.DataFrame, list[float, ...]]
        dataframe with annotations that can be written out as a STAR file and a list of the selected scores
    """

    score_volume = read_mrc(job.output_dir.joinpath(f'{job.tomo_id}_scores.mrc'))
    angle_volume = read_mrc(job.output_dir.joinpath(f'{job.tomo_id}_angles.mrc'))
    angle_list = load_angle_list(
        job.rotation_file,
        sort_angles=version.parse(job.pytom_tm_version_number) > version.parse('0.3.0')
    )

    if tophat_filter:  # constrain the extraction with a tophat filter
        predicted_peaks = predict_tophat_mask(
            score_volume,
            output_path=job.output_dir.joinpath(f'{job.tomo_id}_tophat_filter.svg'),
            n_false_positives=n_false_positives,
            create_plot=create_plot
        )
        score_volume *= predicted_peaks  # multiply with predicted peaks to keep only those

    # apply tomogram mask if provided
    if tomogram_mask_path is not None:
        tomogram_mask = read_mrc(tomogram_mask_path)[
            job.search_origin[0]: job.search_origin[0] + job.search_size[0],
            job.search_origin[1]: job.search_origin[1] + job.search_size[1],
            job.search_origin[2]: job.search_origin[2] + job.search_size[2]
        ]  # mask should be larger than zero in regions of interest!
        score_volume[tomogram_mask <= 0] = 0

    # mask edges of score volume
    score_volume[0: particle_radius_px, :, :] = 0
    score_volume[:, 0: particle_radius_px, :] = 0
    score_volume[:, :, 0: particle_radius_px] = 0
    score_volume[-particle_radius_px:, :, :] = 0
    score_volume[:, -particle_radius_px:, :] = 0
    score_volume[:, :, -particle_radius_px:] = 0

    sigma = job.job_stats['std']
    search_space = job.job_stats['search_space']
    if cut_off is None:
        # formula Rickgauer et al. (2017, eLife): N**(-1) = erfc( theta / ( sigma * sqrt(2) ) ) / 2
        # we need to find theta (i.e. the cut off)
        cut_off = erfcinv((2 * n_false_positives) / search_space) * np.sqrt(2) * sigma
        logging.info(f'cut off for particle extraction: {cut_off}')
    elif cut_off < 0:
        logging.warning('Provided extraction score cut-off is smaller than 0. Changing to 0 as that is smallest '
                        'allowed value.')
        cut_off = 0

    # mask for iteratively selecting peaks
    cut_box = int(particle_radius_px) * 2 + 1
    cut_mask = (spherical_mask(cut_box, particle_radius_px, cut_box // 2) == 0) * 1

    # data for star file
    pixel_size = job.voxel_size
    tomogram_id = job.tomo_id

    data = []
    scores = []

    for _ in tqdm(range(n_particles)):

        ind = np.unravel_index(score_volume.argmax(), score_volume.shape)

        lcc_max = score_volume[ind]

        if lcc_max <= cut_off:
            break

        scores.append(lcc_max)

        # assumed that here also need to multiply with -1
        rotation = convert_euler(
            [-1 * a for a in angle_list[int(angle_volume[ind])]],
            order_in='ZXZ',
            order_out='ZYZ',
            degrees_in=False,
            degrees_out=True
        )

        location = [i + o for i, o in zip(job.search_origin, ind)]

        data.append((
            location[0],  # CoordinateX
            location[1],  # CoordinateY
            location[2],  # CoordinateZ
            rotation[0],  # AngleRot
            rotation[1],  # AngleTilt
            rotation[2],  # AnglePsi
            lcc_max,  # LCCmax
            cut_off,  # Extraction cut off
            sigma,  # Add sigma of template matching search, LCCmax can be divided by sigma to obtain SNR
            pixel_size,  # DetectorPixelSize
            tomogram_id,  # MicrographName
        ))

        # box out the particle
        start = [i - particle_radius_px for i in ind]
        score_volume[
            start[0]: start[0] + cut_box,
            start[1]: start[1] + cut_box,
            start[2]: start[2] + cut_box
        ] *= cut_mask

    output = pd.DataFrame(data, columns=[
        'ptmCoordinateX',
        'ptmCoordinateY',
        'ptmCoordinateZ',
        'ptmAngleRot',
        'ptmAngleTilt',
        'ptmAnglePsi',
        'ptmLCCmax',
        'ptmCutOff',
        'ptmSearchStd',
        'ptmDetectorPixelSize',
        'ptmMicrographName',
    ]), scores

    if plotting_available and create_plot:
        y, bins = np.histogram(scores, bins=20)
        x = (bins[1:] + bins[:-1]) / 2
        hist_step = bins[1] - bins[0]
        # add more starting values for background Gaussian
        x_ext = np.concatenate((np.linspace(x[0] - 5 * hist_step, x[0], 10), x))
        # calculate amplitude of Gaussian and correct for histogram step size
        noise_amplitude = (search_space / (sigma * np.sqrt(2 * np.pi))) * hist_step
        y_background = noise_amplitude * np.exp(- x_ext ** 2 / (2 * sigma ** 2))

        fig, ax = plt.subplots()
        ax.scatter(x, y, label='extracted', marker='o')
        ax.plot(x_ext, y_background, label='background', color='tab:orange')
        ax.axvline(cut_off, color='gray', linestyle='dashed', label='cut-off')
        ax.set_ylim(bottom=0, top=2 * max(y))
        ax.set_ylabel('Occurence')
        ax.set_xlabel(r'${LCC}_{max}$')
        ax.legend()
        plt.tight_layout()
        plt.savefig(
            job.output_dir.joinpath(f'{job.tomo_id}_extraction_graph.svg'),
            dpi=600,
            transparent=False,
            bbox_inches='tight'
        )

    return output



"""
This file name explicitly starts with test00_ to ensure its run first during testing.
Other tests will run jobs on the GPU which keeps the main unittest process lingering on the GPU. When starting the
parallel manager test on a GPU in exclusive process mode, the spawned process from the parallel manager will fail due
the occupation from the other unittests. If the parallel manager is tested first, the spawned process is fully closed
which will allow the remaining test to use the GPU.
"""
import unittest
import pathlib
import numpy as np
import voltools as vt
import multiprocessing
from importlib_resources import files
from pytom_tm.mask import spherical_mask
from pytom_tm.angles import load_angle_list
from pytom_tm.parallel import run_job_parallel
from pytom_tm.tmjob import TMJob
from pytom_tm.io import write_mrc


TOMO_SHAPE = (100, 107, 59)
TEMPLATE_SIZE = 13
LOCATION = (77, 26, 40)
ANGLE_ID = 100
ANGULAR_SEARCH = 'angles_38.53_256.txt'
TEST_DATA_DIR = pathlib.Path(__file__).parent.joinpath('test_data')
TEST_TOMOGRAM = TEST_DATA_DIR.joinpath('tomogram.mrc')
TEST_TEMPLATE = TEST_DATA_DIR.joinpath('template.mrc')
TEST_MASK = TEST_DATA_DIR.joinpath('mask.mrc')


class TestTMJob(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Create template, mask and tomogram
        volume = np.zeros(TOMO_SHAPE, dtype=np.float32)
        template = np.zeros((TEMPLATE_SIZE,) * 3, dtype=np.float32)
        template[3:8, 4:8, 3:7] = 1.
        template[7, 8, 5:7] = 1.
        mask = spherical_mask(TEMPLATE_SIZE, 5, 0.5)
        rotation = load_angle_list(files('pytom_tm.angle_lists').joinpath(ANGULAR_SEARCH))[ANGLE_ID]

        volume[LOCATION[0] - TEMPLATE_SIZE // 2: LOCATION[0] + TEMPLATE_SIZE // 2 + TEMPLATE_SIZE % 2,
               LOCATION[1] - TEMPLATE_SIZE // 2: LOCATION[1] + TEMPLATE_SIZE // 2 + TEMPLATE_SIZE % 2,
               LOCATION[2] - TEMPLATE_SIZE // 2: LOCATION[2] + TEMPLATE_SIZE // 2 + TEMPLATE_SIZE % 2] = vt.transform(
            template,
            rotation=rotation,
            rotation_units='rad',
            rotation_order='rzxz',
            device='cpu'
        )

        # add some noise
        rng = np.random.default_rng(0)
        volume += rng.normal(loc=0, scale=0.1, size=volume.shape)

        TEST_DATA_DIR.mkdir(exist_ok=True)
        write_mrc(TEST_MASK, mask, 1.)
        write_mrc(TEST_TEMPLATE, template, 1.)
        write_mrc(TEST_TOMOGRAM, volume, 1.)

    @classmethod
    def tearDownClass(cls) -> None:
        TEST_MASK.unlink()
        TEST_TEMPLATE.unlink()
        TEST_TOMOGRAM.unlink()
        TEST_DATA_DIR.rmdir()

    def setUp(self):
        self.job = TMJob('0', 10, TEST_TOMOGRAM, TEST_TEMPLATE, TEST_MASK, TEST_DATA_DIR,
                         angle_increment='38.53', voxel_size=1.)

    def test_parallel_breaking(self):
        try:
            _ = run_job_parallel(self.job, volume_splits=(1, 2, 1), gpu_ids=[0, -1], unittest_mute=True)
        except RuntimeError:
            self.assertEqual(len(multiprocessing.active_children()), 0,
                             msg='a process was still lingering after a parallel job with partially invalid resources '
                                 'was started')
        else: # pragma: no cover
            self.fail('This should have given a RuntimeError')

    def test_parallel_manager(self):
        score, angle = run_job_parallel(self.job, volume_splits=(1, 3, 1), gpu_ids=[0])
        ind = np.unravel_index(score.argmax(), score.shape)

        self.assertTrue(score.max() > 0.931, msg='lcc max value lower than expected')
        self.assertEqual(ANGLE_ID, angle[ind])
        self.assertSequenceEqual(LOCATION, ind)



import pathlib
from importlib_resources import files
from scipy.spatial.transform import Rotation


ANGLE_LIST_DIR = files('pytom_tm.angle_lists')
AVAILABLE_ROTATIONAL_SAMPLING = {
    '7.00': ['angles_7.00_45123.txt', 45123],
    '35.76': ['angles_35.76_320.txt', 320],
    '19.95': ['angles_19.95_1944.txt', 1944],
    '90.00': ['angles_90.00_26.txt', 26],
    '18.00': ['angles_18.00_3040.txt', 3040],
    '12.85': ['angles_12.85_7112.txt', 7112],
    '38.53': ['angles_38.53_256.txt', 256],
    '11.00': ['angles_11.00_15192.txt', 15192],
    '17.86': ['angles_17.86_3040.txt', 3040],
    '25.25': ['angles_25.25_980.txt', 980],
    '50.00': ['angles_50.00_100.txt', 100],
    '3.00': ['angles_3.00_553680.txt', 553680],
}
for v in AVAILABLE_ROTATIONAL_SAMPLING.values():
    v[0] = ANGLE_LIST_DIR.joinpath(v[0])


def load_angle_list(file_name: pathlib.Path, sort_angles: bool = True) -> list[tuple[float, float, float]]:
    """Load an angular search list from disk.

    Parameters
    ----------
    file_name: pathlib.Path
        path to text file containing angular search, each line should contain 3 floats of anti-clockwise ZXZ
    sort_angles: bool, default True
        sort the list, using python default angle_list.sort(), sorts first on Z1, then X, then Z2

    Returns
    -------
    angle_list: list[tuple[float, float, float]]
        a list where each element is a tuple of 3 floats containing an anti-clockwise ZXZ Euler rotation in degrees
    """
    with open(str(file_name)) as fstream:
        lines = fstream.readlines()
    angle_list = [tuple(map(float, x.strip().split(' '))) for x in lines]
    if not all([len(a) == 3 for a in angle_list]):
        raise ValueError('Invalid angle file provided, each line should have 3 ZXZ Euler angles!')
    if sort_angles:
        angle_list.sort()  # angle list needs to be sorted otherwise symmetry reduction cannot be used!
    return angle_list


def convert_euler(
        angles: tuple[float, float, float],
        order_in: str = 'ZXZ',
        order_out: str = 'ZXZ',
        degrees_in: bool = True,
        degrees_out: bool = True
) -> tuple[float, float, float]:
    """Convert a single set of Euler angles from one Euler notation to another. This function makes use of
    scipy.spatial.transform.Rotation meaning that capital letters (i.e. ZXZ) specify intrinsic rotations (commonly
    used in cryo-EM) and small letters (i.e. zxz) specific extrinsic rotations.

    Parameters
    ----------
    angles: tuple[float, float, float]
        tuple of three angles
    order_in: str, default 'ZXZ'
        Euler rotation axis of input angles
    order_out: str, default 'ZXZ'
        Euler rotation axis of output angles
    degrees_in: bool, default True
        whether the input angles are in degrees
    degrees_out: bool, default True
        whether the output angles should be in degrees

    Returns
    -------
    output: tuple[float, float, float]
        tuple of three angles
    """
    r = Rotation.from_euler(order_in, angles, degrees=degrees_in)
    return tuple(r.as_euler(order_out, degrees=degrees_out))



import pathlib
import mrcfile
import argparse
import logging
import numpy.typing as npt
import numpy as np
from operator import attrgetter
from typing import Optional, Union


class ParseLogging(argparse.Action):
    """argparse.Action subclass to parse logging parameter from input scripts. Users can set these to info/debug."""
    def __call__(self, parser, namespace, values: str, option_string: Optional[str] = None):
        if not values.upper() in ['INFO', 'DEBUG']:
            parser.error("{0} log got an invalid option, set either to `info` or `debug` ".format(option_string))
        else:
            numeric_level = getattr(logging, values.upper(), None)
            setattr(namespace, self.dest, numeric_level)


class CheckDirExists(argparse.Action):
    """argparse.Action subclass to check if an expected input directory exists."""
    def __call__(self, parser, namespace, values: pathlib.Path, option_string: Optional[str] = None):
        if not values.is_dir():
            parser.error("{0} got a file path that does not exist ".format(option_string))

        setattr(namespace, self.dest, values)


class CheckFileExists(argparse.Action):
    """argparse.Action subclass to check if an expected input file exists."""
    def __call__(self, parser, namespace, values: pathlib.Path, option_string: Optional[str] = None):
        if not values.exists():
            parser.error("{0} got a file path that does not exist ".format(option_string))

        setattr(namespace, self.dest, values)


class LargerThanZero(argparse.Action):
    """argparse.Action subclass to constrain an input value to larger than zero only."""
    def __call__(self, parser, namespace, values: Union[int, float], option_string: Optional[str] = None):
        if values <= .0:
            parser.error("{0} must be larger than 0".format(option_string))

        setattr(namespace, self.dest, values)


class BetweenZeroAndOne(argparse.Action):
    """argparse.Action subclass to constrain an input value to a fraction, i.e. between 0 and 1."""
    def __call__(self, parser, namespace, values: float, option_string: Optional[str] = None):
        if 1. <= values <= .0:
            parser.error("{0} is a fraction and can only range between 0 and 1".format(option_string))

        setattr(namespace, self.dest, values)


class ParseSearch(argparse.Action):
    """argparse.Action subclass to restrict the search area of tomogram to these indices along an axis. Checks that
    these value are larger than zero and that the second value is larger than the first."""
    def __call__(self, parser, namespace, values: list[int, int], option_string: Optional[str] = None):
        if not (0 <= values[0] < values[1]):
            parser.error("{0} start and end indices must be larger than 0 and end must be larger than start".format(
                option_string))

        setattr(namespace, self.dest, values)


class ParseTiltAngles(argparse.Action):
    """argparse.Action subclass to parse tilt_angle info. The input can either be two floats that specify the tilt
    range for a continous wedge model. Alternatively can be a .tlt/.rawtlt file that specifies all the the tilt
    angles of the tilt-series to use for more refined wedge models."""
    def __call__(self, parser, namespace, values: Union[list[str, str], str], option_string: Optional[str] = None):
        if len(values) == 2:  # two wedge angles provided the min and max
            try:
                values = sorted(list(map(float, values)))  # make them floats
                setattr(namespace, self.dest, values)
            except ValueError:
                parser.error("{0} the two arguments provided could not be parsed to floats".format(option_string))
        elif len(values) == 1:
            values = pathlib.Path(values[0])
            if not values.exists() or values.suffix not in ['.tlt', '.rawtlt']:
                parser.error("{0} provided tilt angle file does not exist or does not have the right format".format(
                    option_string))
            setattr(namespace, self.dest, read_tlt_file(values))
        else:
            parser.error("{0} can only take one or two arguments".format(option_string))


class ParseDoseFile(argparse.Action):
    """argparse.Action subclass to parse a txt file contain information on accumulated dose per tilt."""
    def __call__(self, parser, namespace, values: str, option_string: Optional[str] = None):
        file_path = pathlib.Path(values)
        if not file_path.exists():
            parser.error("{0} provided dose accumulation file does not exist".format(option_string))
        allowed_suffixes = ['.txt']
        if file_path.suffix not in allowed_suffixes:
            parser.error("{0}  provided dose accumulation file does not have the right suffix, "
                         "allowed are: {1}".format(option_string, ', '.join(allowed_suffixes)))
        setattr(namespace, self.dest, read_dose_file(file_path))


class ParseDefocusFile(argparse.Action):
    """argparse.Action subclass to read a defocus file, either from IMOD which adheres to their file format,
    or a txt file containing per line the defocus of each tilt."""
    def __call__(self, parser, namespace, values: str, option_string: Optional[str] = None):
        file_path = pathlib.Path(values)
        if not file_path.exists():
            parser.error("{0} provided defocus file does not exist".format(option_string))
        allowed_suffixes = ['.defocus', '.txt']
        if file_path.suffix not in allowed_suffixes:
            parser.error("{0} provided defocus file does not have the right suffix, "
                         "allowed are: {1}".format(option_string, ', '.join(allowed_suffixes)))
        setattr(namespace, self.dest, read_defocus_file(file_path))


class UnequalSpacingError(Exception):
    """Exception for an mrc file that has unequal spacing along the xyz dimensions annotated in its voxel size
    metadata."""
    pass


def write_angle_list(data: npt.NDArray[float], file_name: pathlib.Path, order: tuple[int, int, int] = (0, 2, 1)):
    """Helper function to write angular search list from old PyTom to current module. Order had to be changed as old
    PyTom always stored it as Z1, Z2, X, and here its Z1, X, Z2.

    @todo remove function
    """
    with open(file_name, 'w') as fstream:
        for i in range(data.shape[1]):
            fstream.write(' '.join([str(x) for x in [data[j, i] for j in order]]) + '\n')


def read_mrc_meta_data(file_name: pathlib.Path, permissive: bool = True) -> dict:
    """Read the metadata of provided MRC file path (using mrcfile) and return as dict.

    If the voxel size along the x,y,and z dimensions differs a lot (not within 3 decimals) the function will raise an
    UnequalSpacingError as it could mean template matching on these volumes might not be consistent.

    Parameters
    ----------
    file_name: pathlib.Path
        path to an MRC file
    permissive: bool
        True (default) reads file in permissive mode, setting to False will be more strict with bad MRC headers

    Returns
    -------
    metadata: dict
        a dictionary of the mrc metadata with key 'shape' containing the x,y,z dimensions of the file and key
        'voxel_size' containing the voxel size along x,y,z and dimensions in A units
    """
    meta_data = {}
    with mrcfile.mmap(file_name, permissive=permissive) as mrc:
        meta_data['shape'] = tuple(map(int, attrgetter('nx', 'ny', 'nz')(mrc.header)))
        # allow small numerical inconsistencies in voxel size of MRC headers, sometimes seen in Warp
        if not all(
                [np.round(mrc.voxel_size.x, 3) == np.round(s, 3)
                 for s in attrgetter('y', 'z')(mrc.voxel_size)]
        ):
            raise UnequalSpacingError('Input volume voxel spacing is not identical in each dimension!')
        else:
            if not all([mrc.voxel_size.x == s for s in attrgetter('y', 'z')(mrc.voxel_size)]):
                logging.warning(f'Voxel size annotation in MRC is slightly different between dimensions, '
                                f'namely {mrc.voxel_size}. It might be a tiny numerical inaccuracy, but '
                                f'please ensure this is not problematic.')
            meta_data['voxel_size'] = float(mrc.voxel_size.x)
    return meta_data


def write_mrc(
        file_name: pathlib.Path,
        data: npt.NDArray[float],
        voxel_size: float,
        overwrite: bool = True,
        transpose: bool = True
) -> None:
    """Write data to an MRC file. Data is transposed before writing as pytom internally uses xyz ordering and MRCs
    use zyx.

    Parameters
    ----------
    file_name: pathlib.Path
        path on disk to write the file to
    data: npt.NDArray[float]
        numpy array to write as MRC
    voxel_size: float
        voxel size of array to annotate in MRC header
    overwrite: bool, default True
        True (default) will overwrite current MRC on path, setting to False will error when writing to existing file
    transpose: bool, default True
        True (default) transpose array before writing, setting to False prevents this

    Returns
    -------
    """
    if data.dtype != np.float32:
        logging.warning(f'data for mrc writing is not np.float32 will convert to np.float32')
        data = data.astype(np.float32)
    mrcfile.write(file_name, data.T if transpose else data, voxel_size=voxel_size, overwrite=overwrite)


def read_mrc(
        file_name: pathlib.Path,
        permissive: bool = True,
        transpose: bool = True
) -> npt.NDArray[float]:
    """Read an MRC file from disk. Data is transposed after reading as pytom internally uses xyz ordering and MRCs
    use zyx.

    Parameters
    ----------
    file_name: pathlib.Path
        path to file on disk
    permissive: bool, default True
        True (default) reads file in permissive mode, setting to False will be more strict with bad headers
    transpose: bool, default True
        True (default) transposes the volume after reading, setting to False prevents transpose but probably not a
        good idea when using the functions from this module

    Returns
    -------
    data: npt.NDArray[float]
        returns the MRC data as a numpy array
    """
    with mrcfile.open(file_name, permissive=permissive) as mrc:
        data = np.ascontiguousarray(mrc.data.T) if transpose else mrc.data
    return data


def read_txt_file(file_name: pathlib.Path) -> list[float, ...]:
    """Read a txt file from disk with on each line a single float value.

    Parameters
    ----------
    file_name: pathlib.Path
        file on disk to read

    Returns
    -------
    output: list[float, ...]
        list of floats
    """
    with open(file_name, 'r') as fstream:
        lines = fstream.readlines()
    return list(map(float, [x.strip() for x in lines if not x.isspace()]))


def read_tlt_file(file_name: pathlib.Path) -> list[float, ...]:
    """Read a txt file from disk using read_txt_file(). File is expected to have tilt angles in degrees.

    Parameters
    ----------
    file_name: pathlib.Path
        file on disk to read

    Returns
    -------
    output: list[float, ...]
        list of floats with tilt angles
    """
    return read_txt_file(file_name)


def read_dose_file(file_name: pathlib.Path) -> list[float, ...]:
    """Read a txt file from disk using read_txt_file(). File is expected to have dose accumulation in e-/(A^2).

    Parameters
    ----------
    file_name: pathlib.Path
        file on disk to read

    Returns
    -------
    output: list[float, ...]
        list of floats with accumulated dose
    """
    return read_txt_file(file_name)


def read_imod_defocus_file(file_name: pathlib.Path) -> list[float, ...]:
    """Read an IMOD style defocus file. This function can read version 2 and 3 defocus files. For format
    specification see: https://bio3d.colorado.edu/imod/doc/man/ctfphaseflip.html (section: Defocus File Format).

    Parameters
    ----------
    file_name: pathlib.Path
        file on disk to read

    Returns
    -------
    output: list[float, ...]
        list of floats with defocus (in μm)
    """
    with open(file_name, 'r') as fstream:
        lines = fstream.readlines()
    imod_defocus_version = float(lines[0].strip().split()[5])
    # imod defocus files have the values specified in nm:
    if imod_defocus_version == 2:  # file with one defocus value; data starts on line 0
        return [float(x.strip().split()[4]) * 1e-3 for x in lines]
    elif imod_defocus_version == 3:  # file with astigmatism; line 0 contains metadata that we do not need
        return [(float(x.strip().split()[4]) + float(x.strip().split()[5])) / 2 * 1e-3 for x in lines[1:]]
    else:
        raise ValueError('Invalid IMOD defocus file inversion, can only be 2 or 3.')


def read_defocus_file(file_name: pathlib.Path) -> list[float, ...]:
    """Read a defocus file with values in nm. Output returns defocus in μm.

    Depending on file suffix the function calls:
     - read_imod_defocus_file() for .defocus suffix
     - read_txt_file for .txt suffix

    Parameters
    ----------
    file_name: pathlib.Path
        file on disk to read

    Returns
    -------
    output: list[float, ...]
        list of floats with defocus (in μm)
    """
    if file_name.suffix == '.defocus':
        return read_imod_defocus_file(file_name)
    elif file_name.suffix == '.txt':
        return [x * 1e-3 for x in read_txt_file(file_name)]
    else:
        raise ValueError('Defocus file needs to have format .defocus or .txt')



import argparse
import sys
import pathlib
import logging
import numpy as np
import starfile
from pytom_tm.extract import extract_particles
from pytom_tm.io import (
    LargerThanZero,
    write_mrc,
    read_mrc_meta_data,
    read_mrc,
    CheckFileExists,
    ParseLogging,
    CheckDirExists,
    ParseSearch,
    ParseTiltAngles,
    ParseDoseFile,
    ParseDefocusFile,
    BetweenZeroAndOne,
)
from pytom_tm.tmjob import load_json_to_tmjob


def _parse_argv(argv=None):
    if argv is None:
        return sys.argv[1:]
    return argv


def pytom_create_mask(argv=None):
    from pytom_tm.mask import spherical_mask, ellipsoidal_mask

    argv = _parse_argv(argv)

    parser = argparse.ArgumentParser(
        description="Create a mask for template matching. "
        "-- Marten Chaillet (@McHaillet)"
    )
    parser.add_argument(
        "-b",
        "--box-size",
        type=int,
        required=True,
        action=LargerThanZero,
        help="Shape of square box for the mask.",
    )
    parser.add_argument(
        "-o",
        "--output-file",
        type=pathlib.Path,
        required=False,
        help="Provide path to write output, needs to end in .mrc ."
        "If not provided file is written to current directory in the following format: "
        "./mask_b[box_size]px_r[radius]px.mrc ",
    )
    parser.add_argument(
        "--voxel-size",
        type=float,
        required=False,
        default=1.0,
        action=LargerThanZero,
        help="Provide a voxel size to annotate the MRC (currently not used for any "
        "mask calculation).",
    )
    parser.add_argument(
        "-r",
        "--radius",
        type=float,
        required=True,
        action=LargerThanZero,
        help="Radius of the spherical mask in number of pixels. In case minor1 and "
        "minor2 are provided, this will be the radius of the ellipsoidal mask along "
        "the x-axis.",
    )
    parser.add_argument(
        "--radius-minor1",
        type=float,
        required=False,
        action=LargerThanZero,
        help="Radius of the ellipsoidal mask along the y-axis in number of pixels.",
    )
    parser.add_argument(
        "--radius-minor2",
        type=float,
        required=False,
        action=LargerThanZero,
        help="Radius of the ellipsoidal mask along the z-axis in number of pixels.",
    )
    parser.add_argument(
        "-s",
        "--sigma",
        type=float,
        required=False,
        action=LargerThanZero,
        help="Sigma of gaussian drop-off around the mask edges in number of pixels. "
        "Values in the range from 0.5-1.0 are usually sufficient for tomograms with "
        "20A-10A voxel sizes.",
    )
    argv = _parse_argv(argv)
    args = parser.parse_args(argv)

    # generate mask
    if args.radius_minor1 is not None and args.radius_minor2 is not None:
        mask = ellipsoidal_mask(
            args.box_size,
            args.radius,
            args.radius_minor1,
            args.radius_minor2,
            smooth=args.sigma,
        )
    else:
        mask = spherical_mask(args.box_size, args.radius, smooth=args.sigma)

    # write to disk
    output_path = (
        args.output_file
        if args.output_file is not None
        else (pathlib.Path(f"mask_b{args.box_size}px_r{args.radius}px.mrc"))
    )
    write_mrc(output_path, mask, args.voxel_size)


def pytom_create_template(argv=None):
    from pytom_tm.template import generate_template_from_map

    argv = _parse_argv(argv)
    parser = argparse.ArgumentParser(
        description="Generate template from MRC density. "
        "-- Marten Chaillet (@McHaillet)"
    )
    parser.add_argument(
        "-i",
        "--input-map",
        type=pathlib.Path,
        required=True,
        action=CheckFileExists,
        help="Map to generate template from; MRC file.",
    )
    parser.add_argument(
        "-o",
        "--output-file",
        type=pathlib.Path,
        required=False,
        help="Provide path to write output, needs to end in .mrc . If not provided "
        "file is written to current directory in the following format: "
        "template_{input_map.stem}_{voxel_size}A.mrc",
    )
    parser.add_argument(
        "--input-voxel-size-angstrom",
        type=float,
        required=False,
        action=LargerThanZero,
        help="Voxel size of input map, in Angstrom. If not provided will be read from "
        "MRC input (so make sure it is annotated correctly!).",
    )
    parser.add_argument(
        "--output-voxel-size-angstrom",
        type=float,
        required=True,
        action=LargerThanZero,
        help="Output voxel size of the template, in Angstrom. Needs to be equal to the "
        "voxel size of the tomograms for template matching. Input map will be "
        "downsampled to this spacing.",
    )
    parser.add_argument(
        "--center",
        action="store_true",
        default=False,
        required=False,
        help="Set this flag to automatically center the density in the volume by "
        "measuring the center of mass.",
    )
    parser.add_argument(
        "-c",
        "--ctf-correction",
        action="store_true",
        default=False,
        required=False,
        help="Set this flag to multiply the input map with a CTF. The following "
        "parameters are also important to specify because the defaults might not apply "
        "to your data: --defocus, --amplitude-contrast, --voltage, --Cs.",
    )
    parser.add_argument(
        "-z",
        "--defocus",
        type=float,
        required=False,
        default=3.0,
        help="Defocus in um (negative value is overfocus).",
    )
    parser.add_argument(
        "-a",
        "--amplitude-contrast",
        type=float,
        required=False,
        default=0.08,
        help="Fraction of amplitude contrast in the image ctf.",
    )
    parser.add_argument(
        "-v",
        "--voltage",
        type=float,
        required=False,
        default=300.0,
        help="Acceleration voltage of electrons in keV",
    )
    parser.add_argument(
        "--Cs",
        type=float,
        required=False,
        default=2.7,
        help="Spherical aberration in mm.",
    )
    parser.add_argument(
        "--cut-after-first-zero",
        action="store_true",
        default=False,
        required=False,
        help="Set this flag to cut the CTF after the first zero crossing. Generally "
        "recommended to apply as the simplistic CTF convolution will likely become "
        "inaccurate after this point due to defocus gradients.",
    )
    parser.add_argument(
        "--flip-phase",
        action="store_true",
        default=False,
        required=False,
        help="Set this flag to apply a phase flipped CTF. Only required if the CTF is "
        "modelled beyond the first zero crossing and if the tomograms have been CTF "
        "corrected by phase flipping.",
    )
    parser.add_argument(
        "--low-pass",
        type=float,
        required=False,
        action=LargerThanZero,
        help="Apply a low pass filter to this resolution, in Angstrom. By default a "
        "low pass filter is applied to a resolution of (2 * output_spacing_angstrom) "
        "before downsampling the input volume.",
    )
    parser.add_argument(
        "-b",
        "--box-size",
        type=int,
        required=False,
        action=LargerThanZero,
        help="Specify a desired size for the output box of the template. "
        "Only works if it is larger than the downsampled box size of the input.",
    )
    parser.add_argument(
        "--invert",
        action="store_true",
        default=False,
        required=False,
        help="Multiply template by -1. "
        "WARNING: not needed if ctf with defocus is already applied!",
    ),
    parser.add_argument(
        "-m",
        "--mirror",
        action="store_true",
        default=False,
        required=False,
        help="Mirror the final template before writing to disk.",
    )
    parser.add_argument(
        "--display-filter",
        action="store_true",
        default=False,
        required=False,
        help="Display the combined CTF and low pass filter to the user.",
    )
    parser.add_argument(
        "--log",
        type=str,
        required=False,
        default=20,
        action=ParseLogging,
        help="Can be set to `info` or `debug`",
    )
    args = parser.parse_args(argv)
    logging.basicConfig(level=args.log)

    # set input voxel size and give user warning if it does not match
    # with MRC annotation
    input_data = read_mrc(args.input_map)
    input_meta_data = read_mrc_meta_data(args.input_map)
    if args.input_voxel_size_angstrom is not None:
        if round(args.input_voxel_size_angstrom, 3) != round(
            input_meta_data["voxel_size"], 3
        ):
            logging.warning(
                "Provided voxel size does not match voxel size annotated in input map."
            )
        map_spacing_angstrom = args.input_voxel_size_angstrom
    else:
        map_spacing_angstrom = input_meta_data["voxel_size"]

    # set output path
    output_path = (
        args.output_file
        if args.output_file is not None
        else (
            pathlib.Path(
                f"template_{args.input_map.stem}_{args.output_voxel_size_angstrom}A.mrc"
            )
        )
    )

    if map_spacing_angstrom > args.output_voxel_size_angstrom:
        raise NotImplementedError(
            "It is assumed the input map has smaller voxel size than the output "
            "template."
        )

    ctf_params = None
    if args.ctf_correction:
        ctf_params = {
            "pixel_size": map_spacing_angstrom * 1e-10,
            "defocus": args.defocus * 1e-6,
            "amplitude_contrast": args.amplitude_contrast,
            "voltage": args.voltage * 1e3,
            "spherical_aberration": args.Cs * 1e-3,
            "cut_after_first_zero": args.cut_after_first_zero,
            "flip_phase": args.flip_phase,
        }

    template = generate_template_from_map(
        input_data,
        map_spacing_angstrom,
        args.output_voxel_size_angstrom,
        center=args.center,
        ctf_params=ctf_params,
        filter_to_resolution=args.low_pass,
        output_box_size=args.box_size,
        display_filter=args.display_filter,
    ) * (-1 if args.invert else 1)

    logging.debug(f"shape of template after processing is: {template.shape}")

    write_mrc(
        output_path,
        np.flip(template, axis=0) if args.mirror else template,
        args.output_voxel_size_angstrom,
    )


def estimate_roc(argv=None):
    argv = _parse_argv(argv)
    from pytom_tm.plotting import plist_quality_gaussian_fit

    parser = argparse.ArgumentParser(
        description="Estimate ROC curve from TMJob file. "
        "-- Marten Chaillet (@McHaillet)"
    )
    parser.add_argument(
        "-j",
        "--job-file",
        type=pathlib.Path,
        required=True,
        action=CheckFileExists,
        help="JSON file that contain all data on the template matching job, written "
        "out by pytom_match_template.py in the destination path.",
    )
    parser.add_argument(
        "-n",
        "--number-of-particles",
        type=int,
        required=True,
        action=LargerThanZero,
        help="The number of particles to extract and estimate the ROC on, recommended "
        "is to multiply the expected number of particles by 3.",
    )
    parser.add_argument(
        "-r",
        "--radius-px",
        type=int,
        required=True,
        action=LargerThanZero,
        help="Particle radius in pixels in the tomogram. It is used during extraction "
        "to remove areas around peaks preventing double extraction.",
    )
    parser.add_argument(
        "--bins",
        type=int,
        required=False,
        action=LargerThanZero,
        default=20,
        help="Number of bins for the histogram to fit Gaussians on.",
    )
    parser.add_argument(
        "--gaussian-peak",
        type=int,
        required=False,
        action=LargerThanZero,
        help="Expected index of the histogram peak of the Gaussian fitted to the "
        "particle population.",
    )
    parser.add_argument(
        "--force-peak",
        action="store_true",
        default=False,
        required=False,
        help="Force the particle peak to the provided peak index.",
    )
    parser.add_argument(
        "--crop-plot",
        action="store_true",
        default=False,
        required=False,
        help="Flag to crop the plot relative to the height of the particle population.",
    )
    parser.add_argument(
        "--show-plot",
        action="store_true",
        default=False,
        required=False,
        help="Flag to use a pop-up window for the plot instead of writing it to the "
        "location of the job file.",
    )
    parser.add_argument(
        "--log",
        type=str,
        required=False,
        default=20,
        action=ParseLogging,
        help="Can be set to `info` or `debug`",
    )
    args = parser.parse_args(argv)
    logging.basicConfig(level=args.log)

    template_matching_job = load_json_to_tmjob(args.job_file)
    # Set cut off to -1 to ensure the number of particles gets extracted
    _, lcc_max_values = extract_particles(
        template_matching_job, args.radius_px, args.number_of_particles, cut_off=0
    )

    score_volume = read_mrc(
        template_matching_job.output_dir.joinpath(
            f"{template_matching_job.tomo_id}_scores.mrc"
        )
    )

    plist_quality_gaussian_fit(
        lcc_max_values,
        score_volume,
        args.bins // 2 if args.gaussian_peak is None else args.gaussian_peak,
        force_peak=args.force_peak,
        output_figure_name=(
            None
            if args.show_plot
            else template_matching_job.output_dir.joinpath(
                f"{template_matching_job.tomo_id}_roc.svg"
            )
        ),
        crop_hist=args.crop_plot,
        num_bins=args.bins,
        n_tomograms=1,
    )


def extract_candidates(argv=None):
    argv = _parse_argv(argv)
    parser = argparse.ArgumentParser(
        description="Run candidate extraction. -- Marten Chaillet (@McHaillet)"
    )
    parser.add_argument(
        "-j",
        "--job-file",
        type=pathlib.Path,
        required=True,
        action=CheckFileExists,
        help="JSON file that contain all data on the template matching job, written "
        "out by pytom_match_template.py in the destination path.",
    )
    parser.add_argument(
        "--tomogram-mask",
        type=pathlib.Path,
        required=False,
        action=CheckFileExists,
        help="Here you can provide a mask for the extraction with dimensions equal to "
        "the tomogram. All values in the mask that are smaller or equal to 0 will be "
        "removed, all values larger than 0 are considered regions of interest. It can "
        "be used to extract annotations only within a specific cellular region.",
    )
    parser.add_argument(
        "-n",
        "--number-of-particles",
        type=int,
        required=True,
        action=LargerThanZero,
        help="Maximum number of particles to extract from tomogram.",
    )
    parser.add_argument(
        "--number-of-false-positives",
        type=int,
        required=False,
        action=LargerThanZero,
        help="Number of false positives to determine the false alarm rate. Here one "
        "can increase the recall of the particle of interest at the expense of more "
        "false positives. The default value of 1 is recommended for particles that can "
        "be distinguished well from the background (high specificity).",
        default=1,
    )
    parser.add_argument(
        "-r",
        "--radius-px",
        type=int,
        required=True,
        action=LargerThanZero,
        help="Particle radius in pixels in the tomogram. It is used during extraction "
        "to remove areas around peaks preventing double extraction.",
    )
    parser.add_argument(
        "-c",
        "--cut-off",
        type=float,
        required=False,
        help="Override automated extraction cutoff estimation and instead extract the "
        "number-of-particles down to this LCCmax value. Setting to 0 will keep "
        "extracting until number-of-particles, or until there are no positive values "
        "left in the score map. Values larger than 1 make no sense as the correlation "
        "cannot be higher than 1.",
    )
    parser.add_argument(
        "--tophat-filter",
        action="store_true",
        default=False,
        required=False,
        help="Attempt to filter only sharp correlation peaks with a tophat transform",
    )
    parser.add_argument(
        "--log",
        type=str,
        required=False,
        default=20,
        action=ParseLogging,
        help="Can be set to `info` or `debug`",
    )
    args = parser.parse_args(argv)
    logging.basicConfig(level=args.log)

    # load job and extract particles from the volumes
    job = load_json_to_tmjob(args.job_file)
    df, _ = extract_particles(
        job,
        args.radius_px,
        args.number_of_particles,
        cut_off=args.cut_off,
        n_false_positives=args.number_of_false_positives,
        tomogram_mask_path=args.tomogram_mask,
        tophat_filter=args.tophat_filter,
    )

    # write out as a RELION type starfile
    starfile.write(
        df, job.output_dir.joinpath(f"{job.tomo_id}_particles.star"), overwrite=True
    )


def match_template(argv=None):
    from pytom_tm.tmjob import TMJob
    from pytom_tm.parallel import run_job_parallel

    argv = _parse_argv(argv)
    parser = argparse.ArgumentParser(
        description="Run template matching. -- Marten Chaillet (@McHaillet)"
    )
    parser.add_argument(
        "-t",
        "--template",
        type=pathlib.Path,
        required=True,
        action=CheckFileExists,
        help="Template; MRC file.",
    )
    parser.add_argument(
        "-m",
        "--mask",
        type=pathlib.Path,
        required=True,
        action=CheckFileExists,
        help="Mask with same box size as template; MRC file.",
    )
    parser.add_argument(
        "--non-spherical-mask",
        action="store_true",
        required=False,
        help="Flag to set when the mask is not spherical. It adds the required "
        "computations for non-spherical masks and roughly doubles computation time.",
    )
    parser.add_argument(
        "-v",
        "--tomogram",
        type=pathlib.Path,
        required=True,
        action=CheckFileExists,
        help="Tomographic volume; MRC file.",
    )
    parser.add_argument(
        "-d",
        "--destination",
        type=pathlib.Path,
        required=False,
        default="./",
        action=CheckDirExists,
        help="Folder to store the files produced by template matching.",
    )
    parser.add_argument(
        "-a",
        "--tilt-angles",
        nargs="+",
        type=str,
        required=True,
        action=ParseTiltAngles,
        help="Tilt angles of the tilt-series, either the minimum and maximum values of "
        "the tilts (e.g. --tilt-angles -59.1 60.1) or a .rawtlt/.tlt file with all the "
        "angles (e.g. --tilt-angles tomo101.rawtlt). In case all the tilt angles are "
        "provided a more elaborate Fourier space constraint can be used",
    )
    parser.add_argument(
        "--per-tilt-weighting",
        action="store_true",
        default=False,
        required=False,
        help="Flag to activate per-tilt-weighting, only makes sense if a file with all "
        "tilt angles have been provided. In case not set, while a tilt angle file is "
        "provided, the minimum and maximum tilt angle are used to create a binary "
        "wedge. The base functionality creates a fanned wedge where each tilt is "
        "weighted by cos(tilt_angle). If dose accumulation and CTF parameters are "
        "provided these will all be incorporated in the tilt-weighting.",
    )
    parser.add_argument(
        "--angular-search",
        type=str,
        required=True,
        help="Options are: [7.00, 35.76, 19.95, 90.00, 18.00, "
        "12.85, 38.53, 11.00, 17.86, 25.25, 50.00, 3.00].\n"
        "Alternatively, a .txt file can be provided with three Euler angles "
        "(in radians) per line that define the angular search. "
        "Angle format is ZXZ anti-clockwise (see: "
        "https://www.ccpem.ac.uk/user_help/rotation_conventions.php).",
    )
    parser.add_argument(
        "--z-axis-rotational-symmetry",
        type=int,
        required=False,
        action=LargerThanZero,
        default=1,
        help="Integer value indicating the rotational symmetry of the template around "
        "the z-axis. The length of the rotation search will be shortened through "
        "division by this value. Only works for template symmetry around the z-axis.",
    )
    parser.add_argument(
        "-s",
        "--volume-split",
        nargs=3,
        type=int,
        required=False,
        default=[1, 1, 1],
        help="Split the volume into smaller parts for the search, "
        "can be relevant if the volume does not fit into GPU memory. "
        "Format is x y z, e.g. --volume-split 1 2 1",
    )
    parser.add_argument(
        "--search-x",
        nargs=2,
        type=int,
        required=False,
        action=ParseSearch,
        help="Start and end indices of the search along the x-axis, "
        "e.g. --search-x 10 490 ",
    )
    parser.add_argument(
        "--search-y",
        nargs=2,
        type=int,
        required=False,
        action=ParseSearch,
        help="Start and end indices of the search along the y-axis, "
        "e.g. --search-x 10 490 ",
    )
    parser.add_argument(
        "--search-z",
        nargs=2,
        type=int,
        required=False,
        action=ParseSearch,
        help="Start and end indices of the search along the z-axis, "
        "e.g. --search-x 30 230 ",
    )
    parser.add_argument(
        "--voxel-size-angstrom",
        type=float,
        required=False,
        action=LargerThanZero,
        help="Voxel spacing of tomogram/template in angstrom, if not provided will "
        "try to read from the MRC files. Argument is important for band-pass "
        "filtering!",
    )
    parser.add_argument(
        "--low-pass",
        type=float,
        required=False,
        action=LargerThanZero,
        help="Apply a low-pass filter to the tomogram and template. Generally desired "
        "if the template was already filtered to a certain resolution. "
        "Value is the resolution in A.",
    )
    parser.add_argument(
        "--high-pass",
        type=float,
        required=False,
        action=LargerThanZero,
        help="Apply a high-pass filter to the tomogram and template to reduce "
        "correlation with large low frequency variations. Value is a resolution in A, "
        "e.g. 500 could be appropriate as the CTF is often incorrectly modelled "
        "up to 50nm.",
    )
    parser.add_argument(
        "--dose-accumulation",
        type=str,
        required=False,
        action=ParseDoseFile,
        help="Here you can provide a file that contains the accumulated dose at each "
        "tilt angle, assuming the same ordering of tilts as the tilt angle file. "
        "Format should be a .txt file with on each line a dose value in e-/A2.",
    )
    parser.add_argument(
        "--defocus-file",
        type=str,
        required=False,
        action=ParseDefocusFile,
        help="Here you can provide an IMOD defocus file (version 2 or 3) "
        "or a text file with defocus. The values, together with the other ctf "
        "parameters (amplitude contrast, voltage, spherical abberation), "
        "will be used to create a 3D CTF weighting function. IMPORTANT: if "
        "you provide this, the input template should not be modulated with a CTF "
        "beforehand. Format should be .defocus (IMOD) or .txt, "
        "same ordering as tilt angle list. The .txt file should contain a single "
        "defocus value (in nm) per line.",
    )
    parser.add_argument(
        "--amplitude-contrast",
        type=float,
        required=False,
        action=BetweenZeroAndOne,
        help="Amplitude contrast fraction for CTF.",
    )
    parser.add_argument(
        "--spherical-abberation",
        type=float,
        required=False,
        action=LargerThanZero,
        help="Spherical abberation for CTF in mm.",
    )
    parser.add_argument(
        "--voltage",
        type=float,
        required=False,
        action=LargerThanZero,
        help="Voltage for CTF in keV.",
    )
    parser.add_argument(
        "--spectral-whitening",
        action="store_true",
        default=False,
        required=False,
        help="Calculate a whitening filtering from the power spectrum of the tomogram; "
        "apply it to the tomogram patch and template. Effectively puts more weight on "
        "high resolution features and sharpens the correlation peaks.",
    )
    parser.add_argument(
        "-g",
        "--gpu-ids",
        nargs="+",
        type=int,
        required=True,
        help="GPU indices to run the program on.",
    )
    parser.add_argument(
        "--log",
        type=str,
        required=False,
        default=20,
        action=ParseLogging,
        help="Can be set to `info` or `debug`",
    )
    args = parser.parse_args(argv)
    logging.basicConfig(level=args.log)

    # combine ctf values to ctf_params list of dicts
    ctf_params = None
    if args.defocus_file is not None:
        if (
            args.amplitude_contrast is None
            or args.spherical_abberation is None
            or args.voltage is None
        ):
            raise ValueError(
                "Cannot create 3D CTF weighting because one or multiple of "
                "the required parameters (amplitude-contrast, "
                "spherical-abberation or voltage) is/are missing."
            )
        ctf_params = [
            {
                "defocus": defocus,
                "amplitude": args.amplitude_contrast,
                "voltage": args.voltage,
                "cs": args.spherical_abberation,
            }
            for defocus in args.defocus_file
        ]

    job = TMJob(
        "0",
        args.log,
        args.tomogram,
        args.template,
        args.mask,
        args.destination,
        angle_increment=args.angular_search,
        mask_is_spherical=True
        if args.non_spherical_mask is None
        else (not args.non_spherical_mask),
        tilt_angles=args.tilt_angles,
        tilt_weighting=args.per_tilt_weighting,
        search_x=args.search_x,
        search_y=args.search_y,
        search_z=args.search_z,
        voxel_size=args.voxel_size_angstrom,
        low_pass=args.low_pass,
        high_pass=args.high_pass,
        dose_accumulation=args.dose_accumulation,
        ctf_data=ctf_params,
        whiten_spectrum=args.spectral_whitening,
        rotational_symmetry=args.z_axis_rotational_symmetry,
    )

    score_volume, angle_volume = run_job_parallel(
        job, tuple(args.volume_split), args.gpu_ids
    )

    # set the appropriate headers when writing!
    write_mrc(
        args.destination.joinpath(f"{job.tomo_id}_scores.mrc"),
        score_volume,
        job.voxel_size,
    )
    write_mrc(
        args.destination.joinpath(f"{job.tomo_id}_angles.mrc"),
        angle_volume,
        job.voxel_size,
    )

    # write the job as well
    job.write_to_json(args.destination.joinpath(f"{job.tomo_id}_job.json"))


def merge_stars(argv=None):
    import pandas as pd

    parser = argparse.ArgumentParser(
        description=(
            "Merge multiple star files in the same directory. "
            "-- Marten Chaillet (@McHaillet)"
        )
    )
    parser.add_argument(
        "-i",
        "--input-dir",
        type=pathlib.Path,
        required=False,
        default="./",
        action=CheckDirExists,
        help=(
            "Directory with star files, "
            "script will try to merge all files that end in '.star'."
        ),
    )
    parser.add_argument(
        "-o",
        "--output-file",
        type=pathlib.Path,
        required=False,
        default="./particles.star",
        help="Output star file name.",
    )
    parser.add_argument(
        "--log",
        type=str,
        required=False,
        default=20,
        action=ParseLogging,
        help="Can be set to `info` or `debug`",
    )
    args = parser.parse_args(argv)
    logging.basicConfig(level=args.log)

    files = [f for f in args.input_dir.iterdir() if f.suffix == ".star"]

    if len(files) == 0:
        raise ValueError("No starfiles in directory.")

    logging.info("Concatting and writing star files")

    dataframes = [starfile.read(f) for f in files]

    starfile.write(
        pd.concat(dataframes, ignore_index=True), args.output_file, overwrite=True
    )



import unittest
import numpy as np
from pytom_tm.extract import predict_tophat_mask


class TestExtract(unittest.TestCase):
    def test_predict_tophat_mask(self):
        rng = np.random.default_rng(0)
        volume = rng.normal(loc=0, scale=0.1, size=(50, ) * 3)  # generate random peaks
        tophat_mask = predict_tophat_mask(volume)
        self.assertEqual(tophat_mask.shape, volume.shape, msg='tophat mask should have same size as input')
        self.assertEqual(tophat_mask.dtype, bool, msg='predicted tophat mask should be boolean')



# No imports of pytom_tm outside of the methods
import unittest
from importlib import reload
# Mock out installed dependencies
orig_import = __import__

# skip tests if optional stuff is not installed
SKIP_PLOT = False
try:
    import pytom_tm.plotting
except:
    SKIP_PLOT = True

def module_not_found_mock(missing_name):
    def import_mock(name, *args):
        if name == missing_name:
            raise ModuleNotFoundError(f"No module named '{name}'")
        return orig_import(name, *args)
    return import_mock

def cupy_import_error_mock(name, *args):
    if name == 'cupy':
        raise ImportError("Failed to import cupy")
    return orig_import(name, *args)


class TestMissingDependencies(unittest.TestCase):

    def test_missing_cupy(self):
        # assert working import
        with self.assertNoLogs(level='WARNING'):
            import pytom_tm
        cupy_not_found = module_not_found_mock('cupy')
        # Test missing cupy
        with unittest.mock.patch('builtins.__import__', side_effect=cupy_not_found):
            with self.assertLogs(level='WARNING') as cm:
                reload(pytom_tm)
            self.assertEqual(len(cm.output), 1)
            self.assertIn("cupy installation not found or not functional", cm.output[0])

    def test_broken_cupy(self):
        # assert working import
        with self.assertNoLogs(level='WARNING'):
            import pytom_tm
        # Test cupy ImportError
        with unittest.mock.patch('builtins.__import__', side_effect=cupy_import_error_mock):
            with self.assertLogs(level='WARNING') as cm:
                reload(pytom_tm)
            self.assertEqual(len(cm.output), 1)
            self.assertIn("cupy installation not found or not functional", cm.output[0])

    @unittest.skipIf(SKIP_PLOT, "plotting module not installed")
    def test_missing_matplotlib(self):
        # assert working import
        import pytom_tm

        matplotlib_not_found = module_not_found_mock('matplotlib.pyplot')
        with unittest.mock.patch('builtins.__import__', side_effect=matplotlib_not_found):
            with self.assertRaisesRegex(ModuleNotFoundError, 'matplotlib'):
                # only pyplot is directly imported so this should be tested
                import matplotlib.pyplot as plt
            # force reload 
            # check if we can still import pytom_tm
            reload(pytom_tm)
            
            # check if plotting is indeed disabled after reload
            # (reload is needed to prevent python import caching)
            self.assertFalse(reload(pytom_tm.template).plotting_available)
            self.assertFalse(reload(pytom_tm.extract).plotting_available)
            # assert that importing the plotting module fails completely
            with self.assertRaisesRegex(RuntimeError, "matplotlib and seaborn"):
                reload(pytom_tm.plotting)

    @unittest.skipIf(SKIP_PLOT, "plotting module not installed")
    def test_missing_seaborn(self):
        # assert working import
        import pytom_tm
        
        seaborn_not_found = module_not_found_mock('seaborn')
        with unittest.mock.patch('builtins.__import__', side_effect=seaborn_not_found):
            with self.assertRaisesRegex(ModuleNotFoundError, 'seaborn'):
                import seaborn
            # check if we can still import pytom_tm
            reload(pytom_tm)
            # check if plotting is indeed disabled
            # (reload is needed to prevent python import caching)
            self.assertFalse(reload(pytom_tm.template).plotting_available)
            self.assertFalse(reload(pytom_tm.extract).plotting_available)
            # assert that importing the plotting module fails completely
            with self.assertRaisesRegex(RuntimeError, "matplotlib and seaborn"):
                reload(pytom_tm.plotting)






import numpy.typing as npt
import multiprocessing as mp
import logging
import queue
import time
import contextlib
from multiprocessing.managers import BaseProxy
from functools import reduce
from pytom_tm.tmjob import TMJob
from pytom_tm.utils import mute_stdout_stderr

try:  # new processes need to be spawned in order to set cupy to use the correct GPU
    mp.set_start_method('spawn')
except RuntimeError:
    pass


def gpu_runner(
        gpu_id: int, task_queue: BaseProxy, result_queue: BaseProxy, log_level: int, unittest_mute: bool
) -> None:
    """Start a GPU runner, each runner should be initialized to a multiprocessing.Process() and manage running jobs
    on a single GPU. Each runner will grab jobs from the task_queue and assign jobs to the result_queue once they
    finish. When the task_queue is empty the gpu_runner will stop.

    Parameters
    ----------
    gpu_id: int
        a GPU index to assign to the runner
    task_queue: mp.managers.BaseProxy
        shared queue from multiprocessing with jobs to run
    result_queue: mp.manager.BaseProxy
        shared queue from multiprocessing for finished jobs
    log_level: int
        log level for logging
    unittest_mute: Bool
        optional muting of runner to prevent unittests flooding the terminal, only use for development
    """
    if unittest_mute:
        mute_context = mute_stdout_stderr
    else:
        mute_context = contextlib.nullcontext
    with mute_context():
        logging.basicConfig(level=log_level)
        while True:
            try:
                job = task_queue.get_nowait()
                result_queue.put_nowait(job.start_job(gpu_id, return_volumes=False))
            except queue.Empty:
                break


def run_job_parallel(
        main_job: TMJob, volume_splits: tuple[int, int, int], gpu_ids: list[int, ...], unittest_mute: bool = False,
) -> tuple[npt.NDArray[float], npt.NDArray[float]]:
    """Run a job in parallel over a single or multiple GPUs. If no volume_splits are given the search is parallelized
    by splitting the angular search. If volume_splits are provided the job will first be split by volume,
    if there are still more GPUs available, the subvolume jobs are still further split by angular search.

    Parameters
    ----------
    main_job: pytom_tm.tmjob.TMJob
        a TMJob object from pytom_tm that contains all data for a search
    volume_splits: tuple[int, int, int]
        tuple of len 3 with splits in x, y, and z
    gpu_ids: list[int, ...]
        list of gpu indices to spread jobs over
    unittest_mute: bool, default False
        boolean to mute spawned process terminal output, only set to True for unittesting

    Returns
    -------
    result: tuple[npt.NDArray[float], npt.NDArray[float]]
        the volumes with the LCCmax and angle ids
    """

    n_pieces = reduce(lambda x, y: x * y, volume_splits)
    jobs = []

    # =================== Splitting into subjobs ===============
    if n_pieces == 1:
        if len(gpu_ids) > 1:  # split rotation search

            jobs = main_job.split_rotation_search(len(gpu_ids))

        else:  # we just run the whole tomo on a single gpu

            jobs.append(main_job)

    elif n_pieces > 1:

        rotation_split_factor = len(gpu_ids) // n_pieces

        if rotation_split_factor >= 2:  # we can split the rotation search for the subvolumes

            for j in main_job.split_volume_search(volume_splits):

                jobs += j.split_rotation_search(rotation_split_factor)

        else:  # only split the subvolume search

            jobs = main_job.split_volume_search(volume_splits)

    else:
        raise ValueError('Invalid number of pieces in split volume')

    # ================== Execution of jobs =========================
    if len(jobs) == 1:

        return main_job.start_job(gpu_ids[0], return_volumes=True)

    elif len(jobs) >= len(gpu_ids):

        results = []

        with mp.Manager() as manager:
            task_queue = manager.Queue()  # the list of tasks where processes can get there next task from
            result_queue = manager.Queue()  # this will accumulate results from the processes

            [task_queue.put_nowait(j) for j in jobs]  # put all tasks

            # set the processes and start them!
            procs = [mp.Process(target=gpu_runner, args=(
                g, task_queue, result_queue, main_job.log_level, unittest_mute)) for g in gpu_ids]
            [p.start() for p in procs]

            while True:
                while not result_queue.empty():
                    results.append(result_queue.get_nowait())

                if len(results) == len(jobs):  # its done if all the results from the spawn were send back
                    logging.debug('Got all results from the child processes')
                    break

                for p in procs:  # if one of the processes is no longer alive and has a failed exit we should error
                    if not p.is_alive() and p.exitcode == 1:  # to prevent a deadlock
                        [x.terminate() for x in procs]  # kill all spawned processes if something broke
                        raise RuntimeError('One or more of the processes stopped unexpectedly.')

                time.sleep(1)

            [p.join() for p in procs]
            logging.debug('Terminated the processes')

        # merge split jobs; pass along the list of stats to annotate them in main_job
        return main_job.merge_sub_jobs(stats=results)

    else:
        ValueError('For some reason there are more gpu_ids than split job, this should never happen.')



import numpy as np
import numpy.typing as npt
from typing import Optional


def spherical_mask(
        box_size: int,
        radius: float,
        smooth: Optional[float] = None,
        cutoff_sd: int = 3,
        center: Optional[float] = None
) -> npt.NDArray[float]:
    """Wrapper around ellipsoidal_mask() to create a spherical mask with just a single radius.

    Parameters
    ----------
    box_size: int
        box size of the mask, equal in each dimension
    radius: float
        radius of sphere
    smooth: Optional[float], default None
        sigma (float relative to number of pixels) of gaussian falloff around mask
    cutoff_sd: int, default 3
        how many standard deviations of the Gaussian falloff to include, default of 3 is a good choice
    center: Optional[float], default None
        alternative center for the mask, default is (size - 1) / 2

    Returns
    -------
    mask: npt.NDArray[float]
        3D numpy array with mask at center
    """
    return ellipsoidal_mask(box_size, radius, radius, radius, smooth, cutoff_sd=cutoff_sd, center=center)


def ellipsoidal_mask(
        box_size: int,
        major: float,
        minor1: float,
        minor2: float,
        smooth: Optional[float] = None,
        cutoff_sd: int = 3,
        center: Optional[float] = None
) -> npt.NDArray[float]:
    """Create an ellipsoidal mask in the specified square box. Ellipsoid is defined by 3 radius on x,y, and z axis.

    Parameters
    ----------
    box_size: int
        box size of the mask, equal in each dimension
    major: float
        radius of ellipsoid in x
    minor1: float
        radius of ellipsoid in y
    minor2: float
        radius of ellipsoid in z
    smooth: Optional[float], default None
        sigma (float relative to number of pixels) of gaussian falloff around mask
    cutoff_sd: int, default 3
        how many standard deviations of the Gaussian falloff to include, default of 3 is a good choice
    center: Optional[float], default None
        alternative center for the mask, default is (size - 1) / 2

    Returns
    -------
    mask: npt.NDArray[float]
        3D numpy array with mask at center
    """
    if not all([box_size > 0, major > 0, minor1 > 0, minor2 > 0]):
        raise ValueError('Invalid input for mask creation: box_size or radii are <= 0')

    center = (box_size - 1) / 2 if center is None else center
    x, y, z = (np.arange(box_size) - center,
               np.arange(box_size) - center,
               np.arange(box_size) - center)

    # use broadcasting
    r = np.sqrt(((x / major)**2)[:, np.newaxis, np.newaxis] +
                ((y / minor1)**2)[:, np.newaxis] +
                (z / minor2)**2).astype(np.float32)

    if smooth is not None:

        if not all([smooth >= 0, cutoff_sd >= 0]):
            raise ValueError('Invalid input for mask smoothing: smooth or sd cutoff are <= 0')

        r[r <= 1] = 1
        sigma = (smooth / ((major + minor1 + minor2) / 3))
        mask = np.exp(-1 * ((r - 1) / sigma) ** 2)
        mask[mask <= np.exp(-cutoff_sd**2/2.)] = 0

    else:

        mask = np.zeros_like(r)
        mask[r <= 1] = 1.

    return mask



import unittest
from shutil import which
from contextlib import redirect_stdout
from io import StringIO
from pytom_tm import entry_points

# (command line function, function in entry_points file)
ENTRY_POINTS_TO_TEST = [
    ("pytom_create_mask.py", "pytom_create_mask"),
    ("pytom_create_template.py", "pytom_create_template"),
    ("pytom_match_template.py", "match_template"),
    ("pytom_extract_candidates.py", "extract_candidates"),
    ("pytom_merge_stars.py",  "merge_stars"),
    ]
# Test if optional dependencies are installed
try:
    from pytom_tm import plotting
except:
    pass
else:
    ENTRY_POINTS_TO_TEST.append(("pytom_estimate_roc.py", "estimate_roc"))

class TestEntryPoints(unittest.TestCase):
    def test_entry_points_exist(self):
        for cli, fname in ENTRY_POINTS_TO_TEST:
            # test the command line function can be found
            self.assertIsNotNone(which(cli))
            # assert the entry_point be called with -h and exit cleanly
            # catch stdout to prevent shell polution
            func = getattr(entry_points, fname)
            dump = StringIO()
            with self.assertRaises(SystemExit) as ex, redirect_stdout(dump):
                func([cli, '-h'])
            dump.close()
            # check if the system return code is 0 (success)
            self.assertEqual(ex.exception.code, 0)



import unittest
import pathlib
import numpy as np
import voltools as vt
import mrcfile
from importlib_resources import files
from pytom_tm.mask import spherical_mask
from pytom_tm.angles import load_angle_list
from pytom_tm.tmjob import TMJob, TMJobError, load_json_to_tmjob
from pytom_tm.io import read_mrc, write_mrc, UnequalSpacingError
from pytom_tm.extract import extract_particles
from test_weights import CTF_PARAMS, ACCUMULATED_DOSE, TILT_ANGLES


TOMO_SHAPE = (100, 107, 59)
TEMPLATE_SIZE = 13
LOCATION = (77, 26, 40)
ANGLE_ID = 100
ANGULAR_SEARCH = 'angles_38.53_256.txt'
TEST_DATA_DIR = pathlib.Path(__file__).parent.joinpath('test_data')
TEST_TOMOGRAM = TEST_DATA_DIR.joinpath('tomogram.mrc')
TEST_EXTRACTION_MASK_OUTSIDE = TEST_DATA_DIR.joinpath('extraction_mask_outside.mrc')
TEST_EXTRACTION_MASK_INSIDE = TEST_DATA_DIR.joinpath('extraction_mask_inside.mrc')
TEST_TEMPLATE = TEST_DATA_DIR.joinpath('template.mrc')
TEST_TEMPLATE_UNEQUAL_SPACING = TEST_DATA_DIR.joinpath('template_unequal_spacing.mrc')
TEST_TEMPLATE_WRONG_VOXEL_SIZE = TEST_DATA_DIR.joinpath('template_voxel_error_test.mrc')
TEST_MASK = TEST_DATA_DIR.joinpath('mask.mrc')
TEST_SCORES = TEST_DATA_DIR.joinpath('tomogram_scores.mrc')
TEST_ANGLES = TEST_DATA_DIR.joinpath('tomogram_angles.mrc')
TEST_CUSTOM_ANGULAR_SEARCH = TEST_DATA_DIR.joinpath('custom_angular_search.txt')
TEST_WHITENING_FILTER = TEST_DATA_DIR.joinpath('tomogram_whitening_filter.npy')
TEST_JOB_JSON = TEST_DATA_DIR.joinpath('tomogram_job.json')
TEST_JOB_JSON_WHITENING = TEST_DATA_DIR.joinpath('tomogram_job_whitening.json')


class TestTMJob(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Create template, mask and tomogram
        volume = np.zeros(TOMO_SHAPE, dtype=np.float32)
        template = np.zeros((TEMPLATE_SIZE,) * 3, dtype=np.float32)
        template[3:8, 4:8, 3:7] = 1.
        template[7, 8, 5:7] = 1.
        mask = spherical_mask(TEMPLATE_SIZE, 5, 0.5)
        rotation = load_angle_list(files('pytom_tm.angle_lists').joinpath(ANGULAR_SEARCH))[ANGLE_ID]

        volume[LOCATION[0] - TEMPLATE_SIZE // 2: LOCATION[0] + TEMPLATE_SIZE // 2 + TEMPLATE_SIZE % 2,
               LOCATION[1] - TEMPLATE_SIZE // 2: LOCATION[1] + TEMPLATE_SIZE // 2 + TEMPLATE_SIZE % 2,
               LOCATION[2] - TEMPLATE_SIZE // 2: LOCATION[2] + TEMPLATE_SIZE // 2 + TEMPLATE_SIZE % 2] = vt.transform(
            template,
            rotation=rotation,
            rotation_units='rad',
            rotation_order='rzxz',
            device='cpu'
        )

        # add some noise
        rng = np.random.default_rng(0)
        volume += rng.normal(loc=0, scale=0.1, size=volume.shape)

        # extraction mask
        extraction_mask_outside = np.zeros(TOMO_SHAPE, dtype=np.float32)
        extraction_mask_outside[20:40, 60:80, 10:30] = 1
        extraction_mask_inside = np.zeros(TOMO_SHAPE, dtype=np.float32)
        extraction_mask_inside[70:90, 15:35, 30:50] = 1

        TEST_DATA_DIR.mkdir(exist_ok=True)
        write_mrc(TEST_EXTRACTION_MASK_OUTSIDE, extraction_mask_outside, 1.)
        write_mrc(TEST_EXTRACTION_MASK_INSIDE, extraction_mask_inside, 1.)
        write_mrc(TEST_MASK, mask, 1.)
        write_mrc(TEST_TEMPLATE, template, 1.)
        write_mrc(TEST_TEMPLATE_WRONG_VOXEL_SIZE, template, 1.5)
        mrcfile.write(TEST_TEMPLATE_UNEQUAL_SPACING, template, voxel_size=(1.5, 1., 2.), overwrite=True)
        write_mrc(TEST_TOMOGRAM, volume, 1.)

        # do a run without splitting to compare against
        job = TMJob(
            '0',
            10,
            TEST_TOMOGRAM,
            TEST_TEMPLATE,
            TEST_MASK,
            TEST_DATA_DIR,
            angle_increment='38.53',
            voxel_size=1.
        )
        score, angle = job.start_job(0, return_volumes=True)
        write_mrc(
            TEST_SCORES,
            score,
            job.voxel_size
        )
        write_mrc(
            TEST_ANGLES,
            angle,
            job.voxel_size
        )
        job.write_to_json(TEST_JOB_JSON)

        np.savetxt(TEST_CUSTOM_ANGULAR_SEARCH, np.random.rand(100, 3))

        # create job with spectrum whitening
        job = TMJob('0', 10, TEST_TOMOGRAM, TEST_TEMPLATE, TEST_MASK, TEST_DATA_DIR,
                    angle_increment='90.00', voxel_size=1., whiten_spectrum=True)
        job.write_to_json(TEST_JOB_JSON_WHITENING)

    @classmethod
    def tearDownClass(cls) -> None:
        TEST_MASK.unlink()
        TEST_EXTRACTION_MASK_OUTSIDE.unlink()
        TEST_EXTRACTION_MASK_INSIDE.unlink()
        TEST_TEMPLATE.unlink()
        TEST_TEMPLATE_UNEQUAL_SPACING.unlink()
        TEST_TEMPLATE_WRONG_VOXEL_SIZE.unlink()
        TEST_TOMOGRAM.unlink()
        TEST_SCORES.unlink()
        TEST_ANGLES.unlink()
        TEST_CUSTOM_ANGULAR_SEARCH.unlink()
        # the whitening filter might not exist if the job with spectrum whitening failed, so the unlinking needs to
        # allow this (with missing_ok=True) to ensure clean up of the test directory
        TEST_WHITENING_FILTER.unlink(missing_ok=True)
        TEST_JOB_JSON.unlink()
        TEST_JOB_JSON_WHITENING.unlink()
        TEST_DATA_DIR.rmdir()

    def setUp(self):
        self.job = TMJob('0', 10, TEST_TOMOGRAM, TEST_TEMPLATE, TEST_MASK, TEST_DATA_DIR,
                         angle_increment='38.53', voxel_size=1.)

    def test_tm_job_errors(self):
        with self.assertRaises(ValueError, msg='Different voxel size in tomogram and template and no voxel size '
                                               'provided should raise a ValueError'):
            TMJob('0', 10, TEST_TOMOGRAM, TEST_TEMPLATE_WRONG_VOXEL_SIZE, TEST_MASK, TEST_DATA_DIR)

        with self.assertRaises(UnequalSpacingError,
                               msg='Unequal spacing should raise specific Error'):
            TMJob('0', 10, TEST_TOMOGRAM, TEST_TEMPLATE_UNEQUAL_SPACING, TEST_MASK, TEST_DATA_DIR)

        # test searches raise correct errors
        for param in ['search_x', 'search_y', 'search_z']:
            with self.assertRaises(ValueError, msg='Invalid start index in search should raise ValueError'):
                TMJob('0', 10, TEST_TOMOGRAM, TEST_TEMPLATE, TEST_MASK, TEST_DATA_DIR, voxel_size=1.,
                      **{param: [-10, 100]})
            with self.assertRaises(ValueError, msg='Invalid start index in search should raise ValueError'):
                TMJob('0', 10, TEST_TOMOGRAM, TEST_TEMPLATE, TEST_MASK, TEST_DATA_DIR, voxel_size=1.,
                      **{param: [110, 130]})
            with self.assertRaises(ValueError, msg='Invalid end index in search should raise ValueError'):
                TMJob('0', 10, TEST_TOMOGRAM, TEST_TEMPLATE, TEST_MASK, TEST_DATA_DIR, voxel_size=1.,
                      **{param: [0, 120]})

        with self.assertRaises(TMJobError, msg='Angular increment for which no default list is available should raise '
                                               'a TMJobError'):
            TMJob('0', 10, TEST_TOMOGRAM, TEST_TEMPLATE, TEST_MASK, TEST_DATA_DIR, voxel_size=1.,
                  angle_increment='5.41')

    def test_tm_job_copy(self):
        copy = self.job.copy()
        self.assertIsNot(
            self.job, copy,
            msg='Copying the job should create a new object.'
        )
        self.assertEqual(
            TOMO_SHAPE, copy.tomo_shape,
            msg='Tomogram shape not correct in job, perhaps transpose issue?'
        )

    def test_tm_job_weighting_options(self):
        # run with all options
        job = TMJob('0', 10, TEST_TOMOGRAM, TEST_TEMPLATE, TEST_MASK, TEST_DATA_DIR,
                    angle_increment='90.00', voxel_size=1., low_pass=10, high_pass=100,
                    dose_accumulation=ACCUMULATED_DOSE, ctf_data=CTF_PARAMS, tilt_angles=TILT_ANGLES,
                    whiten_spectrum=True, tilt_weighting=True)
        score, angle = job.start_job(0, return_volumes=True)
        self.assertEqual(score.shape, job.tomo_shape, msg='TMJob with all options failed')

        # run with only tilt weighting (in test_weights different options are tested for create_wedge)
        job = TMJob('0', 10, TEST_TOMOGRAM, TEST_TEMPLATE, TEST_MASK, TEST_DATA_DIR,
                    angle_increment='90.00', voxel_size=1.,
                    dose_accumulation=ACCUMULATED_DOSE, ctf_data=CTF_PARAMS, tilt_angles=TILT_ANGLES,
                    tilt_weighting=True)
        score, angle = job.start_job(0, return_volumes=True)
        self.assertEqual(score.shape, job.tomo_shape, msg='TMJob with only wedge creation failed')

        # run with only bandpass (in test_weights bandpass option are tested)
        job = TMJob('0', 10, TEST_TOMOGRAM, TEST_TEMPLATE, TEST_MASK, TEST_DATA_DIR,
                    angle_increment='90.00', voxel_size=1., low_pass=10, high_pass=100)
        score, angle = job.start_job(0, return_volumes=True)
        self.assertEqual(score.shape, job.tomo_shape, msg='TMJob with only band-pass failed')

        # run with only spectrum whitening (in test_weights the whitening filter is tested
        job = TMJob('0', 10, TEST_TOMOGRAM, TEST_TEMPLATE, TEST_MASK, TEST_DATA_DIR,
                    angle_increment='90.00', voxel_size=1., whiten_spectrum=True)
        score, angle = job.start_job(0, return_volumes=True)
        self.assertEqual(score.shape, job.tomo_shape, msg='TMJob with only whitening filter failed')

        # load the whitening filter from previous job to compare against
        whitening_filter = np.load(TEST_WHITENING_FILTER)
        job = TMJob('0', 10, TEST_TOMOGRAM, TEST_TEMPLATE, TEST_MASK, TEST_DATA_DIR,
                    angle_increment='90.00', voxel_size=1., whiten_spectrum=True, search_y=[10, 90])
        new_whitening_filter = np.load(TEST_WHITENING_FILTER)
        self.assertNotEqual(whitening_filter.shape, new_whitening_filter.shape,
                            msg='After reducing the search region along the largest dimension the whitening filter '
                                'should have less sampling points')
        self.assertEqual(new_whitening_filter.shape, (max(job.search_size) // 2 + 1, ),
                         msg='The whitening filter does not have the expected size, it should be equal (x // 2) + 1, '
                             'where x is the largest dimension of the search box.')

        # TMJob with none of these weighting options is tested in all other runs in this file.

    def test_load_json_to_tmjob(self):
        # check base job loading
        job = load_json_to_tmjob(TEST_JOB_JSON)
        self.assertIsInstance(job, TMJob, msg='TMJob could not be properly loaded from disk.')

        # check job loading and preventing whitening filter recalculation
        with self.assertNoLogs(level='INFO'):
            _ = load_json_to_tmjob(TEST_JOB_JSON_WHITENING, load_for_extraction=True)
        with self.assertLogs(level='INFO') as cm:
            _ = load_json_to_tmjob(TEST_JOB_JSON_WHITENING, load_for_extraction=False)
        self.assertEqual(len(cm.output), 1)
        self.assertIn('Estimating whitening filter...', cm.output[0])

    def test_custom_angular_search(self):
        job = TMJob('0', 10, TEST_TOMOGRAM, TEST_TEMPLATE, TEST_MASK, TEST_DATA_DIR,
                    angle_increment=TEST_CUSTOM_ANGULAR_SEARCH, voxel_size=1.)
        self.assertEqual(job.rotation_file, TEST_CUSTOM_ANGULAR_SEARCH, msg='Passing a custom angular search file to '
                                                                            'TMJob failed.')

    def test_tm_job_split_volume(self):
        with self.assertRaises(RuntimeError, msg='Splitting the volume into smaller boxes than the template should '
                                                 'raise an error.'):
            self.job.split_volume_search((10, 3, 2))

        sub_jobs = self.job.split_volume_search((2, 3, 2))
        stats = []
        for x in sub_jobs:
            stats.append(x.start_job(0))
            job_scores = TEST_DATA_DIR.joinpath(f'tomogram_scores_{x.job_key}.mrc')
            job_angles = TEST_DATA_DIR.joinpath(f'tomogram_angles_{x.job_key}.mrc')
            self.assertTrue(
                job_scores.exists(),
                msg='Expected output from job does not exist.'
            )
            self.assertTrue(
                job_angles.exists(),
                msg='Expected output from job does not exist.'
            )
        score, angle = self.job.merge_sub_jobs(stats)
        ind = np.unravel_index(score.argmax(), score.shape)

        self.assertTrue(score.max() > 0.931, msg='lcc max value lower than expected')
        self.assertEqual(ANGLE_ID, angle[ind])
        self.assertSequenceEqual(LOCATION, ind)

        # Small difference in the edge regions of the split dimension. This is because the cross correlation function
        # is not well defined in the boundary area, only a small part of the template is correlated here (and we are
        # not really interested in it). Probably the inaccuracy in this area becomes more apparent when splitting
        # into subvolumes due to a smaller number of sampling points in Fourier space.
        ok_region = slice(TEMPLATE_SIZE // 2, -TEMPLATE_SIZE // 2)
        score_diff = np.abs( 
            score[ok_region, ok_region, ok_region] -
            read_mrc(TEST_SCORES)[ok_region, ok_region, ok_region]
            ).sum() 
        angle_diff = np.abs(
            angle[ok_region, ok_region, ok_region] -
            read_mrc(TEST_ANGLES)[ok_region, ok_region, ok_region]
            ).sum()
        self.assertAlmostEqual(score_diff, 0, places=1, msg='score diff should not be larger than 0.01')
        self.assertAlmostEqual(angle_diff, 0, places=1, msg='angle diff should not change')

        # get search statistics before and after splitting
        split_stats = self.job.job_stats
        reference_stats = load_json_to_tmjob(TEST_JOB_JSON).job_stats
        self.assertEqual(split_stats['search_space'], reference_stats['search_space'],
                         msg='Search space should remain identical upon subvolume splitting.')
        self.assertAlmostEqual(split_stats['std'], reference_stats['std'], places=3,
                               msg='Standard deviation over template matching with subvolume splitting should be '
                                   'almost identical.')

    def test_splitting_with_offsets(self):
        # check if subjobs have correct offsets for the main job, the last sub job will have the largest errors
        job = TMJob('0', 10, TEST_TOMOGRAM, TEST_TEMPLATE, TEST_MASK, TEST_DATA_DIR,
                    angle_increment='38.53', voxel_size=1., search_x=[9, 90], search_y=[25, 102], search_z=[19, 54])
        # split along each dimension and get only the last sub job
        last_sub_job = job.split_volume_search((2, 3, 2))[-1]
        final_size = [
            i + j - TEMPLATE_SIZE // 2
            for i, j
            in zip(last_sub_job.whole_start, last_sub_job.search_size)
        ]
        self.assertEqual(final_size, job.search_size,
                         msg='the last subjobs start position plus its size (minus template overhang) should equal '
                             'the search size of the main job')

    def test_tm_job_split_angles(self):
        sub_jobs = self.job.split_rotation_search(3)
        stats = []
        for x in sub_jobs:
            stats.append(x.start_job(0))
            job_scores = TEST_DATA_DIR.joinpath(f'tomogram_scores_{x.job_key}.mrc')
            job_angles = TEST_DATA_DIR.joinpath(f'tomogram_angles_{x.job_key}.mrc')
            self.assertTrue(
                job_scores.exists(),
                msg='Expected output from job does not exist.'
            )
            self.assertTrue(
                job_angles.exists(),
                msg='Expected output from job does not exist.'
            )
        score, angle = self.job.merge_sub_jobs(stats)
        ind = np.unravel_index(score.argmax(), score.shape)

        self.assertTrue(score.max() > 0.931, msg='lcc max value lower than expected')
        self.assertEqual(ANGLE_ID, angle[ind])
        self.assertSequenceEqual(LOCATION, ind)

        self.assertTrue(np.abs(score - read_mrc(TEST_SCORES)).sum() == 0,
                        msg='split rotation search should be identical')
        self.assertTrue(np.abs(angle - read_mrc(TEST_ANGLES)).sum() == 0,
                        msg='split rotation search should be identical')

        # get search statistics before and after splitting
        split_stats = self.job.job_stats
        reference_stats = load_json_to_tmjob(TEST_JOB_JSON).job_stats
        self.assertEqual(split_stats['search_space'], reference_stats['search_space'],
                         msg='Search space should remain identical upon angular search splitting.')
        self.assertAlmostEqual(split_stats['std'], reference_stats['std'], places=6,
                               msg='Standard deviation of template matching with angular search split should be almost '
                                   'identical.')

    def test_extraction(self):
        _ = self.job.start_job(0, return_volumes=True)

        # extract particles after running the job
        df, scores = extract_particles(self.job, 5, 100, create_plot=False)
        self.assertNotEqual(len(scores), 0, msg='Here we expect to get some annotations.')

        # test extraction mask that does not cover the particle
        df, scores = extract_particles(self.job, 5, 100,
                                       tomogram_mask_path=TEST_EXTRACTION_MASK_OUTSIDE,
                                       create_plot=False)
        self.assertEqual(len(scores), 0, msg='Length of returned list should be 0 after applying mask where the '
                                             'object is not in the region of interest.')

        # test mask that covers the particle
        df, scores = extract_particles(self.job, 5, 100,
                                       tomogram_mask_path=TEST_EXTRACTION_MASK_INSIDE,
                                       create_plot=False)
        self.assertNotEqual(len(scores), 0, msg='We expected a detected particle with a extraction mask that '
                                                'covers the object.')



from __future__ import annotations
from importlib import metadata
from packaging import version
import pathlib
import copy
import numpy as np
import numpy.typing as npt
import json
import logging
from typing import Optional, Union
from functools import reduce
from scipy.fft import next_fast_len, rfftn, irfftn
from pytom_tm.angles import AVAILABLE_ROTATIONAL_SAMPLING, load_angle_list
from pytom_tm.matching import TemplateMatchingGPU
from pytom_tm.weights import create_wedge, power_spectrum_profile, profile_to_weighting, create_gaussian_band_pass
from pytom_tm.io import read_mrc_meta_data, read_mrc, write_mrc, UnequalSpacingError
from pytom_tm import __version__ as PYTOM_TM_VERSION


def load_json_to_tmjob(file_name: pathlib.Path, load_for_extraction: bool = True) -> TMJob:
    """Load a previous job that was stored with TMJob.write_to_json().

    Parameters
    ----------
    file_name: pathlib.Path
        path to TMJob json file
    load_for_extraction: bool, default True
        whether a finished job is loaded form disk for extraction, default is True as this function is currently only
        called for pytom_extract_candidates and pytom_estimate_roc which run on previously finished jobs

    Returns
    -------
    job: TMJob
        initialized TMJob
    """
    with open(file_name, 'r') as fstream:
        data = json.load(fstream)

    job = TMJob(
        data['job_key'],
        data['log_level'],
        pathlib.Path(data['tomogram']),
        pathlib.Path(data['template']),
        pathlib.Path(data['mask']),
        pathlib.Path(data['output_dir']),
        mask_is_spherical=data['mask_is_spherical'],
        tilt_angles=data['tilt_angles'],
        tilt_weighting=data['tilt_weighting'],
        search_x=data['search_x'],
        search_y=data['search_y'],
        search_z=data['search_z'],
        voxel_size=data['voxel_size'],
        low_pass=data['low_pass'],
        # Use 'get' for backwards compatibility
        high_pass=data.get('high_pass', None),
        dose_accumulation=data.get('dose_accumulation', None),
        ctf_data=data.get('ctf_data', None),
        whiten_spectrum=data.get('whiten_spectrum', False),
        rotational_symmetry=data.get('rotational_symmetry', 1),
        # if version number is not in the .json, it must be 0.3.0 or older
        pytom_tm_version_number=data.get('pytom_tm_version_number', '0.3.0'),
        job_loaded_for_extraction=load_for_extraction,
    )
    job.rotation_file = pathlib.Path(data['rotation_file'])
    job.whole_start = data['whole_start']
    job.sub_start = data['sub_start']
    job.sub_step = data['sub_step']
    job.n_rotations = data['n_rotations']
    job.start_slice = data['start_slice']
    job.steps_slice = data['steps_slice']
    job.job_stats = data['job_stats']
    return job


class TMJobError(Exception):
    """TMJob Exception with provided message."""
    def __init__(self, message):
        # Call the base class constructor with the parameters it needs
        super().__init__(message)


class TMJob:
    def __init__(
            self,
            job_key: str,
            log_level: int,
            tomogram: pathlib.Path,
            template: pathlib.Path,
            mask: pathlib.Path,
            output_dir: pathlib.Path,
            angle_increment: str = '7.00',
            mask_is_spherical: bool = True,
            tilt_angles: Optional[list[float, ...]] = None,
            tilt_weighting: bool = False,
            search_x: Optional[list[int, int]] = None,
            search_y: Optional[list[int, int]] = None,
            search_z: Optional[list[int, int]] = None,
            voxel_size: Optional[float] = None,
            low_pass: Optional[float] = None,
            high_pass: Optional[float] = None,
            dose_accumulation: Optional[list[float, ...]] = None,
            ctf_data: Optional[list[dict, ...]] = None,
            whiten_spectrum: bool = False,
            rotational_symmetry: int = 1,
            pytom_tm_version_number: str = PYTOM_TM_VERSION,
            job_loaded_for_extraction: bool = False,
    ):
        """
        Parameters
        ----------
        job_key: str
            job identifier
        log_level: int
            log level for logging module
        tomogram: pathlib.Path
            path to tomogram MRC
        template: pathlib.Path
            path to template MRC
        mask: pathlib.Path
            path to mask MRC
        output_dir: pathlib.Path
            path to output directory
        angle_increment: str, default '7.00'
            angular increment of template search
        mask_is_spherical: bool, default True
            whether template mask is spherical, reduces computation complexity
        tilt_angles: Optional[list[float, ...]], default None
            tilt angles of tilt-series used to reconstruct tomogram, if only two floats will be used to generate a
            continuous wedge model
        tilt_weighting: bool, default False
            use advanced tilt weighting options, can be supplemented with CTF parameters and accumulated dose
        search_x: Optional[list[int, int]], default None
            restrict tomogram search region along the x-axis
        search_y: Optional[list[int, int]], default None
            restrict tomogram search region along the y-axis
        search_z: Optional[list[int, int]], default None
            restrict tomogram search region along the z-axis
        voxel_size: Optional[float], default None
            voxel size of tomogram and template (in A) if not provided will be read from template/tomogram MRCs
        low_pass: Optional[float], default None
            optional low-pass filter (resolution in A) to apply to tomogram and template
        high_pass: Optional[float], default None
            optional high-pass filter (resolution in A) to apply to tomogram and template
        dose_accumulation: Optional[list[float, ...]], default None
            list with dose accumulation per tilt image
        ctf_data: Optional[list[dict, ...]], default None
            list of dictionaries with CTF parameters per tilt image, see pytom_tm.weight.create_ctf() for parameter
            definition
        whiten_spectrum: bool, default False
            whether to apply spectrum whitening
        rotational_symmetry: int, default 1
            specify a rotational symmetry around the z-axis, is only valid if the symmetry axis of the template is
            aligned with the z-axis
        pytom_tm_version_number: str, default current version
            a string with the version number of pytom_tm for backward compatibility
        job_loaded_for_extraction: bool, default False
            flag to set for finished template matching jobs that are loaded back for extraction, it prevents
            recalculation of the whitening filter which is unnecessary at this stage
        """
        self.mask = mask
        self.mask_is_spherical = mask_is_spherical
        self.output_dir = output_dir

        self.tomogram = tomogram
        self.template = template
        self.tomo_id = self.tomogram.stem

        try:
            meta_data_tomo = read_mrc_meta_data(self.tomogram)
        except UnequalSpacingError:  # add information that the problem is the tomogram
            raise UnequalSpacingError('Input tomogram voxel spacing is not equal in each dimension!')

        try:
            meta_data_template = read_mrc_meta_data(self.template)
        except UnequalSpacingError:  # add information that the problem is the template
            raise UnequalSpacingError('Input template voxel spacing is not equal in each dimension!')

        self.tomo_shape = meta_data_tomo['shape']
        self.template_shape = meta_data_template['shape']

        if voxel_size is not None:
            if voxel_size <= 0:
                raise ValueError('Invalid voxel size provided, smaller or equal to zero.')
            self.voxel_size = voxel_size
            if (  # allow tiny numerical differences that are not relevant for template matching
                    round(self.voxel_size, 3) != round(meta_data_tomo['voxel_size'], 3) or
                    round(self.voxel_size, 3) != round(meta_data_template['voxel_size'], 3)
            ):
                logging.debug(f"provided {self.voxel_size} tomogram {meta_data_tomo['voxel_size']} "
                              f"template {meta_data_template['voxel_size']}")
                print('WARNING: Provided voxel size does not match voxel size annotated in tomogram/template mrc.')
        elif (round(meta_data_tomo['voxel_size'], 3) == round(meta_data_template['voxel_size'], 3) and
              meta_data_tomo['voxel_size'] > 0):
            self.voxel_size = round(meta_data_tomo['voxel_size'], 3)
        else:
            raise ValueError('Voxel size could not be assigned, either a mismatch between tomogram and template or'
                             ' annotated as 0.')

        search_origin = [x[0] if x is not None else 0 for x in (search_x, search_y, search_z)]
        # Check if tomogram origin is valid
        if all([0 <= x < y for x, y in zip(search_origin, self.tomo_shape)]):
            self.search_origin = search_origin
        else:
            raise ValueError('Invalid input provided for search origin of tomogram.')

        # if end not valid raise and error
        search_end = []
        for x, s in zip([search_x, search_y, search_z], self.tomo_shape):
            if x is not None:
                if not x[1] <= s:
                    raise ValueError('One of search end indices is larger than the tomogram dimension.')
                search_end.append(x[1])
            else:
                search_end.append(s)
        self.search_size = [end - start for end, start in zip(search_end, self.search_origin)]

        logging.debug(f'origin, size = {self.search_origin}, {self.search_size}')

        self.whole_start = None
        # For the main job these are always [0,0,0] and self.search_size, for sub_jobs these will differ from
        # self.search_origin and self.search_size. The main job only uses them to calculate the search_volume_roi for
        # statistics. Sub jobs also use these to extract and place back the relevant region in the master job.
        self.sub_start, self.sub_step = [0, 0, 0], self.search_size.copy()

        # Rotation parameters
        self.start_slice = 0
        self.steps_slice = 1
        self.rotational_symmetry = rotational_symmetry
        try:
            self.rotation_file = AVAILABLE_ROTATIONAL_SAMPLING[angle_increment][0]
            self.n_rotations = AVAILABLE_ROTATIONAL_SAMPLING[angle_increment][1]
        except KeyError:
            possible_file_path = pathlib.Path(angle_increment)
            if possible_file_path.exists() and possible_file_path.suffix == '.txt':
                logging.info('Custom file provided for the angular search. Checking if it can be read...')
                # load_angle_list will throw an error if it does not encounter three inputs per line
                self.n_rotations = len(load_angle_list(possible_file_path))
                self.rotation_file = possible_file_path
            else:
                raise TMJobError('Invalid angular search provided.')

        # missing wedge
        self.tilt_angles = tilt_angles
        self.tilt_weighting = tilt_weighting
        # set the band-pass resolution shells
        self.low_pass = low_pass
        self.high_pass = high_pass

        # set dose and ctf
        self.dose_accumulation = dose_accumulation
        self.ctf_data = ctf_data
        self.whiten_spectrum = whiten_spectrum
        self.whitening_filter = self.output_dir.joinpath(f'{self.tomo_id}_whitening_filter.npy')
        if self.whiten_spectrum and not job_loaded_for_extraction:
            logging.info('Estimating whitening filter...')
            weights = 1 / np.sqrt(
                power_spectrum_profile(
                    read_mrc(self.tomogram)[
                        self.search_origin[0]: self.search_origin[0] + self.search_size[0],
                        self.search_origin[1]: self.search_origin[1] + self.search_size[1],
                        self.search_origin[2]: self.search_origin[2] + self.search_size[2]
                    ]
                )
            )
            weights /= weights.max()  # scale to 1
            np.save(self.whitening_filter, weights)

        # Job details
        self.job_key = job_key
        self.leader = None  # the job that spawned this job
        self.sub_jobs = []  # if this job had no sub jobs it should be executed

        # dict to keep track of job statistics
        self.job_stats = None

        self.log_level = log_level

        # version number of the job
        self.pytom_tm_version_number = pytom_tm_version_number

    def copy(self) -> TMJob:
        """Create a copy of the TMJob

        Returns
        -------
        job: TMJob
            copied TMJob instance
        """
        return copy.deepcopy(self)

    def write_to_json(self, file_name: pathlib.Path) -> None:
        """Write job to .json file.

        Parameters
        ----------
        file_name: pathlib.Path
            path to the output file
        """
        d = self.__dict__.copy()
        d.pop('sub_jobs')
        d.pop('search_origin')
        d.pop('search_size')
        d['search_x'] = [self.search_origin[0], self.search_origin[0] + self.search_size[0]]
        d['search_y'] = [self.search_origin[1], self.search_origin[1] + self.search_size[1]]
        d['search_z'] = [self.search_origin[2], self.search_origin[2] + self.search_size[2]]
        for key, value in d.items():
            if isinstance(value, pathlib.Path):
                d[key] = str(value)
        with open(file_name, 'w') as fstream:
            json.dump(d, fstream, indent=4)

    def split_rotation_search(self, n: int) -> list[TMJob, ...]:
        """Split the search into sub_jobs by dividing the rotations. Sub jobs will obtain the key
        self.job_key + str(i) when looping over range(n).

        Parameters
        ----------
        n: int
            number of times to split the angular search

        Returns
        -------
        sub_jobs: list[TMJob, ...]
            a list of TMJobs that were split from self, the jobs are also assigned as the TMJob.sub_jobs attribute
        """
        if len(self.sub_jobs) > 0:
            raise TMJobError('Could not further split this job as it already has subjobs assigned!')

        sub_jobs = []
        for i in range(n):
            new_job = self.copy()
            new_job.start_slice = i
            new_job.steps_slice = n
            new_job.leader = self.job_key
            new_job.job_key = self.job_key + str(i)
            sub_jobs.append(new_job)

        self.sub_jobs = sub_jobs

        return self.sub_jobs

    def split_volume_search(self, split: tuple[int, int, int]) -> list[TMJob, ...]:
        """Split the search into sub_jobs by dividing into subvolumes. Final number of subvolumes is obtained by
        multiplying all the split together, e.g. (2, 2, 1) results in 4 subvolumes. Sub jobs will obtain the key
        self.job_key + str(i) when looping over range(n).

        The sub jobs search area of the full tomogram is defined by: new_job.search_origin and new_job.search_size.
        They are used when loading the search volume from the full tomogram.

        The attribute new_job.whole_start defines how the volume maps back to the score volume of the parent job
        (which can be a different size than the tomogram when the search is restricted along x, y or z).

        Finally, new_job.sub_start and new_job.sub_step, extract the score and angle map without the template
        overhang from the subvolume.

        Parameters
        ----------
        split: tuple[int, int, int]
            tuple that defines how many times the search volume should be split into subvolumes along each axis

        Returns
        -------
        sub_jobs: list[TMJob, ...]
            a list of TMJobs that were split from self, the jobs are also assigned as the TMJob.sub_jobs attribute
        """
        if len(self.sub_jobs) > 0:
            raise TMJobError('Could not further split this job as it already has subjobs assigned!')

        # size of sub-volumes after splitting
        split_size = tuple([x // y for x, y in zip(self.search_size, split)])

        # shape of template for overhang
        template_shape = self.template_shape  # tranposed

        # check if valid
        if any([x > y for x, y in zip(template_shape, split_size)]):
            raise RuntimeError("Not big enough volume to split!")

        # size of subvolume with overhang (underscore indicates not final tuple)
        _size = tuple([x + y for x, y in zip(template_shape, split_size)])

        number_of_pieces = reduce((lambda x, y: x * y), split)

        sub_jobs = []

        for i in range(number_of_pieces):
            stride_z, stride_y = split[0] * split[1], split[0]
            inc_z, inc_y, inc_x = i // stride_z, (i % stride_z) // stride_y, i % stride_y

            _start = tuple([
                -(template_shape[0] // 2) + self.search_origin[0] + inc_x * split_size[0],
                -(template_shape[1] // 2) + self.search_origin[1] + inc_y * split_size[1],
                -(template_shape[2] // 2) + self.search_origin[2] + inc_z * split_size[2]
            ])

            _end = tuple([_start[j] + _size[j] for j in range(len(_start))])

            # if start location is smaller than search origin we need to reset it to search origin
            start = tuple([o if s < o else s for s, o in zip(_start, self.search_origin)])
            # if end location is larger than origin + search size it needs to be set to origin + search size
            end = tuple([s + o if e > s + o else e for e, s, o in zip(_end, self.search_size, self.search_origin)])
            size = tuple([end[j] - start[j] for j in range(len(start))])

            # for reassembling the result
            whole_start = [0, ] * 3  # whole_start and sub_start should also be job attributes
            sub_start = [0, ] * 3
            sub_step = list(split_size)
            if start[0] != self.search_origin[0]:
                whole_start[0] = start[0] + template_shape[0] // 2 - self.search_origin[0]
                sub_start[0] = template_shape[0] // 2
            if start[1] != self.search_origin[1]:
                whole_start[1] = start[1] + template_shape[1] // 2 - self.search_origin[1]
                sub_start[1] = template_shape[1] // 2
            if start[2] != self.search_origin[2]:
                whole_start[2] = start[2] + template_shape[2] // 2 - self.search_origin[2]
                sub_start[2] = template_shape[2] // 2
            if end[0] == self.search_origin[0] + self.search_size[0]:
                sub_step[0] = size[0] - sub_start[0]
            if end[1] == self.search_origin[1] + self.search_size[1]:
                sub_step[1] = size[1] - sub_start[1]
            if end[2] == self.search_origin[2] + self.search_size[2]:
                sub_step[2] = size[2] - sub_start[2]

            # create a split volume job
            new_job = self.copy()
            new_job.leader = self.job_key
            new_job.job_key = self.job_key + str(i)
            new_job.search_origin = (start[0], start[1], start[2])
            new_job.search_size = (size[0], size[1], size[2])
            new_job.whole_start = tuple(whole_start)
            new_job.sub_start = tuple(sub_start)
            new_job.sub_step = tuple(sub_step)
            new_job.split_size = split_size
            sub_jobs.append(new_job)

        self.sub_jobs = sub_jobs

        return self.sub_jobs

    def merge_sub_jobs(self, stats: Optional[list[dict, ...]] = None) -> tuple[npt.NDArray[float], npt.NDArray[float]]:
        """Merge the sub jobs present in self.sub_jobs together to create the final output score and angle maps.

        Parameters
        ----------
        stats: Optional[list[dict, ...]], default None
            optional list of sub job statistics to merge together

        Returns
        -------
        output: tuple[npt.NDArray[float], npt.NDArray[float]]
            the merged score and angle maps from the subjobs
        """
        if len(self.sub_jobs) == 0:
            # read the volumes, remove them and return them
            score_file, angle_file = (
                self.output_dir.joinpath(f'{self.tomo_id}_scores_{self.job_key}.mrc'),
                self.output_dir.joinpath(f'{self.tomo_id}_angles_{self.job_key}.mrc')
            )
            result = (
                read_mrc(score_file),
                read_mrc(angle_file)
            )
            (score_file.unlink(), angle_file.unlink())
            return result

        if stats is not None:
            search_space = sum([s['search_space'] for s in stats])
            variance = sum([s['variance'] for s in stats]) / len(stats)
            self.job_stats = {
                'search_space': search_space,
                'variance': variance,
                'std': np.sqrt(variance)
            }

        is_subvolume_split = np.all(np.array([x.start_slice for x in self.sub_jobs]) == 0)

        score_volumes, angle_volumes = [], []
        for x in self.sub_jobs:
            result = x.merge_sub_jobs()
            score_volumes.append(result[0])
            angle_volumes.append(result[1])

        if not is_subvolume_split:
            scores, angles = np.zeros_like(score_volumes[0]) - 1., np.zeros_like(angle_volumes[0]) - 1.
            for s, a in zip(score_volumes, angle_volumes):
                angles = np.where(s > scores, a, angles)
                # prevents race condition due to slicing
                angles = np.where(s == scores, np.minimum(a, angles), angles)
                scores = np.where(s > scores, s, scores)
        else:
            scores, angles = (
                np.zeros(self.search_size, dtype=np.float32),
                np.zeros(self.search_size, dtype=np.float32)
            )
            for job, s, a in zip(self.sub_jobs, score_volumes, angle_volumes):
                sub_scores = s[
                             job.sub_start[0]: job.sub_start[0] + job.sub_step[0],
                             job.sub_start[1]: job.sub_start[1] + job.sub_step[1],
                             job.sub_start[2]: job.sub_start[2] + job.sub_step[2]]
                sub_angles = a[
                             job.sub_start[0]: job.sub_start[0] + job.sub_step[0],
                             job.sub_start[1]: job.sub_start[1] + job.sub_step[1],
                             job.sub_start[2]: job.sub_start[2] + job.sub_step[2]]
                # Then the corrected sub part needs to be placed back into the full volume
                scores[
                    job.whole_start[0]: job.whole_start[0] + sub_scores.shape[0],
                    job.whole_start[1]: job.whole_start[1] + sub_scores.shape[1],
                    job.whole_start[2]: job.whole_start[2] + sub_scores.shape[2]
                ] = sub_scores
                angles[
                    job.whole_start[0]: job.whole_start[0] + sub_scores.shape[0],
                    job.whole_start[1]: job.whole_start[1] + sub_scores.shape[1],
                    job.whole_start[2]: job.whole_start[2] + sub_scores.shape[2]
                ] = sub_angles
        return scores, angles

    def start_job(
            self,
            gpu_id: int,
            return_volumes: bool = False
    ) -> Union[tuple[npt.NDArray[float], npt.NDArray[float]], dict]:
        """Run this template matching job on the specified GPU. Search statistics of the job will always be assigned
        to the self.job_stats.

        Parameters
        ----------
        gpu_id: int
            index of the GPU to run the job on
        return_volumes: bool, default False
            False (default) does not return volumes but instead writes them to disk, set to True to instead directly
            return the score and angle volumes

        Returns
        -------
        output: Union[tuple[npt.NDArray[float], npt.NDArray[float]], dict]
            when volumes are returned the output consists of two numpy arrays (score and angle map), when no volumes
            are returned the output consists of a dictionary with search statistics
        """
        # next fast fft len
        logging.debug(f'Next fast fft shape: {tuple([next_fast_len(s, real=True) for s in self.search_size])}')
        search_volume = np.zeros(tuple([next_fast_len(s, real=True) for s in self.search_size]), dtype=np.float32)

        # load the (sub)volume
        search_volume[:self.search_size[0],
                      :self.search_size[1],
                      :self.search_size[2]] = np.ascontiguousarray(
            read_mrc(self.tomogram)[self.search_origin[0]: self.search_origin[0] + self.search_size[0],
                                    self.search_origin[1]: self.search_origin[1] + self.search_size[1],
                                    self.search_origin[2]: self.search_origin[2] + self.search_size[2]]
        )

        # load template and mask
        template, mask = (
            read_mrc(self.template),
            read_mrc(self.mask)
        )
        # apply mask directly to prevent any wedge convolution with weird edges
        template *= mask

        # init tomogram and template weighting
        tomo_filter, template_wedge = 1, 1
        # first generate bandpass filters
        if not (self.low_pass is None and self.high_pass is None):
            tomo_filter *= create_gaussian_band_pass(
                    search_volume.shape,
                    self.voxel_size,
                    self.low_pass,
                    self.high_pass
            ).astype(np.float32)
            template_wedge *= create_gaussian_band_pass(
                self.template_shape,
                self.voxel_size,
                self.low_pass,
                self.high_pass
            ).astype(np.float32)

        # then multiply with optional whitening filters
        if self.whiten_spectrum:
            tomo_filter *= profile_to_weighting(
                np.load(self.whitening_filter),
                search_volume.shape
            ).astype(np.float32)
            template_wedge *= profile_to_weighting(
                np.load(self.whitening_filter),
                self.template_shape
            ).astype(np.float32)

        # if tilt angles are provided we can create wedge filters
        if self.tilt_angles is not None:
            # for the tomogram a binary wedge is generated to explicitly set the missing wedge region to 0
            tomo_filter *= create_wedge(
                search_volume.shape,
                self.tilt_angles,
                self.voxel_size,
                cut_off_radius=1.,
                angles_in_degrees=True,
                tilt_weighting=False
            ).astype(np.float32)
            # for the template a binary or per-tilt-weighted wedge is generated depending on the options
            template_wedge *= create_wedge(
                self.template_shape,
                self.tilt_angles,
                self.voxel_size,
                cut_off_radius=1.,
                angles_in_degrees=True,
                tilt_weighting=self.tilt_weighting,
                accumulated_dose_per_tilt=self.dose_accumulation,
                ctf_params_per_tilt=self.ctf_data
            ).astype(np.float32)

        # apply the optional band pass and whitening filter to the search region
        search_volume = np.real(irfftn(rfftn(search_volume) * tomo_filter, s=search_volume.shape))

        # load rotation search
        angle_ids = list(range(
            self.start_slice,
            int(np.ceil(self.n_rotations / self.rotational_symmetry)),
            self.steps_slice
        ))
        angle_list = load_angle_list(
            self.rotation_file,
            sort_angles=version.parse(self.pytom_tm_version_number) > version.parse('0.3.0')
        )[slice(
            self.start_slice,
            int(np.ceil(self.n_rotations / self.rotational_symmetry)),
            self.steps_slice
        )]

        # slices for relevant part for job statistics
        search_volume_roi = (
            slice(self.sub_start[0], self.sub_start[0] + self.sub_step[0]),
            slice(self.sub_start[1], self.sub_start[1] + self.sub_step[1]),
            slice(self.sub_start[2], self.sub_start[2] + self.sub_step[2])
        )

        tm = TemplateMatchingGPU(
            job_id=self.job_key,
            device_id=gpu_id,
            volume=search_volume,
            template=template,
            mask=mask,
            angle_list=angle_list,
            angle_ids=angle_ids,
            mask_is_spherical=self.mask_is_spherical,
            wedge=template_wedge,
            stats_roi=search_volume_roi
        )
        results = tm.run()
        score_volume = results[0][:self.search_size[0], :self.search_size[1], :self.search_size[2]]
        angle_volume = results[1][:self.search_size[0], :self.search_size[1], :self.search_size[2]]
        self.job_stats = results[2]

        del tm  # delete the template matching plan

        if return_volumes:
            return score_volume, angle_volume
        else:  # otherwise write them out with job_key
            write_mrc(
                self.output_dir.joinpath(f'{self.tomo_id}_scores_{self.job_key}.mrc'),
                score_volume,
                self.voxel_size
            )
            write_mrc(
                self.output_dir.joinpath(f'{self.tomo_id}_angles_{self.job_key}.mrc'),
                angle_volume,
                self.voxel_size
            )
            return self.job_stats



"""
Functions in this file are cpu/gpu agnostic.
"""
import numpy.typing as npt
import cupy.typing as cpt
from typing import Optional, Union


def mean_under_mask(
        data: Union[npt.NDArray[float], cpt.NDArray[float]],
        mask: Union[npt.NDArray[float], cpt.NDArray[float]],
        mask_weight: Optional[float] = None
) -> Union[float, cpt.NDArray[float]]:
    """Calculate mean of array in the mask region.

    data and mask can be cupy or numpy arrays.

    Parameters
    ----------
    data: Union[npt.NDArray[float], cpt.NDArray[float]]
        input array
    mask: Union[npt.NDArray[float], cpt.NDArray[float]]
        input mask, same dimensions as data
    mask_weight: Optional[float], default None
        optional weight of mask, if not provided mask.sum() is used to determine weight

    Returns
    -------
    output: Union[float, cpt.NDArray[float]]
        mean of data in the region of the mask
    """
    output = (data * mask).sum() / (mask_weight if mask_weight is not None else mask.sum())
    return output


def std_under_mask(
        data: Union[npt.NDArray[float], cpt.NDArray[float]],
        mask: Union[npt.NDArray[float], cpt.NDArray[float]],
        mean: float,
        mask_weight: Optional[float] = None
) -> Union[float, cpt.NDArray[float]]:
    """Calculate standard deviation of array in the mask region. Uses mean_under_mask() to calculate the mean of
    data**2 within the mask.

    data and mask can be cupy or numpy arrays.

    Parameters
    ----------
    data: Union[npt.NDArray[float], cpt.NDArray[float]]
        input array
    mask: Union[npt.NDArray[float], cpt.NDArray[float]]
        input mask, same dimensions as data
    mean: float
        mean of array in masked region
    mask_weight: Optional[float], default None
        optional weight of mask, if not provided mask.sum() is used to determine weight

    Returns
    -------
    output: Union[float, cpt.NDArray[float]]
        standard deviation of data in the region of the mask
    """
    output = (mean_under_mask(data ** 2, mask, mask_weight=mask_weight) - mean ** 2) ** 0.5
    return output


def normalise(
        data: Union[npt.NDArray[float], cpt.NDArray[float]],
        mask: Optional[Union[npt.NDArray[float], cpt.NDArray[float]]] = None,
        mask_weight: Optional[float] = None
) -> Union[npt.NDArray[float], cpt.NDArray[float]]:
    """Normalise array by subtracting mean and dividing by standard deviation. If a mask is provided the array is
    normalised with the mean and std calculated within the mask.

    data and mask can be cupy or numpy arrays.

    Parameters
    ----------
    data: Union[npt.NDArray[float], cpt.NDArray[float]]
        input array to normalise
    mask: Optional[Union[npt.NDArray[float], cpt.NDArray[float]]], default None
        optional mask to normalise with mean and std in masked region
    mask_weight: Optional[float], default None
        optional float specifying mask weight, if not provided mask.sum() is used

    Returns
    -------
    output: Union[npt.NDArray[float], cpt.NDArray[float]]
        normalised array
    """
    if mask is None:
        mean, std = data.mean(), data.std()
    else:
        mean = mean_under_mask(data, mask, mask_weight=mask_weight)
        std = std_under_mask(data, mask, mean, mask_weight=mask_weight)
    output = (data - mean) / std
    return output


def normalised_cross_correlation(
        data1: Union[npt.NDArray[float], cpt.NDArray[float]],
        data2: Union[npt.NDArray[float], cpt.NDArray[float]],
        mask: Optional[Union[npt.NDArray[float], cpt.NDArray[float]]] = None
) -> Union[float, cpt.NDArray[float]]:
    """Calculate normalised cross correlation between two arrays. Optionally only in a masked region.

    data1, data2, and mask can be cupy or numpy arrays.

    Parameters
    ----------
    data1: Union[npt.NDArray[float], cpt.NDArray[float]]
        first array for correlation
    data2: Union[npt.NDArray[float], cpt.NDArray[float]]
        second array for correlation
    mask: Optional[Union[npt.NDArray[float], cpt.NDArray[float]]], default None
        optional mask to calculate the correlation under

    Returns
    -------
    output: Union[float, cpt.NDArray[float]]
        normalised cross correlation between the arrays
    """
    if mask is None:
        output = (normalise(data1) * normalise(data2)).sum() / data1.size
    else:
        output = (normalise(data1, mask) * mask * normalise(data2, mask)).sum() / mask.sum()
    return output



import unittest
import voltools as vt
import numpy as np
from pytom_tm.matching import TemplateMatchingGPU
from importlib_resources import files
from pytom_tm.mask import spherical_mask
from pytom_tm.angles import load_angle_list


class TestTM(unittest.TestCase):
    def setUp(self):
        self.t_size = 12
        self.volume = np.zeros((100, ) * 3, dtype=float)
        self.template = np.zeros((self.t_size, ) * 3, dtype=float)
        self.template[3:8, 4:8, 3:7] = 1.
        self.mask = spherical_mask(self.t_size, 5, 0.5)
        self.gpu_id = 'gpu:0'
        self.angles = load_angle_list(files('pytom_tm.angle_lists').joinpath('angles_38.53_256.txt'))

    def test_search(self):
        angle_id = 100
        rotation = self.angles[angle_id]
        loc = (77, 26, 40)
        self.volume[loc[0] - self.t_size // 2: loc[0] + self.t_size // 2,
                    loc[1] - self.t_size // 2: loc[1] + self.t_size // 2,
                    loc[2] - self.t_size // 2: loc[2] + self.t_size // 2] = vt.transform(
            self.template,
            rotation=rotation,
            rotation_units='rad',
            rotation_order='rzxz',
            device='cpu'
        )

        tm = TemplateMatchingGPU(0, 0, self.volume, self.template, self.mask, self.angles, list(range(len(
            self.angles))))
        score_volume, angle_volume, stats = tm.run()

        ind = np.unravel_index(score_volume.argmax(), self.volume.shape)
        self.assertTrue(score_volume.max() > 0.99, msg='lcc max value lower than expected')
        self.assertEqual(angle_id, angle_volume[ind])
        self.assertSequenceEqual(loc, ind)
        self.assertEqual(stats['search_space'], 256000000, msg='Search space should exactly equal this value')
        self.assertAlmostEqual(stats['std'], 0.005175, places=5,
                               msg='Standard deviation of the search should be almost equal')



import cupy as cp
import cupy.typing as cpt
import numpy.typing as npt
import voltools as vt
import gc
from typing import Optional
from cupyx.scipy.fft import rfftn, irfftn, fftshift
from tqdm import tqdm
from pytom_tm.correlation import mean_under_mask, std_under_mask
from packaging import version


class TemplateMatchingPlan:
    def __init__(
            self,
            volume: npt.NDArray[float],
            template: npt.NDArray[float],
            mask: npt.NDArray[float],
            device_id: int,
            wedge: Optional[npt.NDArray[float]] = None
    ):
        """Initialize a template matching plan. All the necessary cupy arrays will be allocated on the GPU.

        Parameters
        ----------
        volume: npt.NDArray[float]
            3D numpy array representing the search tomogram
        template: npt.NDArray[float]
            3D numpy array representing the template for the search, a square box of size sx
        mask: npt.NDArray[float]
            3D numpy array representing the mask for the search, same dimensions as template
        device_id: int
            GPU device id to load arrays on
        wedge: Optional[npt.NDArray[float]], default None
            3D numpy array that contains the Fourier space weighting for the template, it should be in Fourier
            reduced form, with dimensions (sx, sx, sx // 2 + 1)
        """
        # Search volume + and fft transform plan for the volume
        self.volume = cp.asarray(volume, dtype=cp.float32, order='C')
        self.volume_rft = rfftn(self.volume)
        # Explicit fft plan is no longer necessary as cupy generates a plan behind the scene which leads to
        # comparable timings

        # Array for storing local standard deviations
        self.std_volume = cp.zeros_like(volume, dtype=cp.float32)

        # Data for the mask
        self.mask = cp.asarray(mask, dtype=cp.float32, order='C')
        self.mask_texture = vt.StaticVolume(self.mask, interpolation='filt_bspline', device=f'gpu:{device_id}')
        self.mask_padded = cp.zeros_like(self.volume).astype(cp.float32)
        self.mask_weight = self.mask.sum()  # weight of the mask

        # Init template data
        self.template = cp.asarray(template, dtype=cp.float32, order='C')
        self.template_texture = vt.StaticVolume(self.template, interpolation='filt_bspline', device=f'gpu:{device_id}')
        self.template_padded = cp.zeros_like(self.volume)

        # fourier binary wedge weight for the template
        self.wedge = cp.asarray(wedge, order='C', dtype=cp.float32) if wedge is not None else None

        # Initialize result volumes
        self.ccc_map = cp.zeros_like(self.volume)
        self.scores = cp.ones_like(self.volume)*-1000
        self.angles = cp.ones_like(self.volume)*-1000

        # wait for stream to complete the work
        cp.cuda.stream.get_current_stream().synchronize()

    def clean(self) -> None:
        """Remove all stored cupy arrays from the GPU's memory pool."""
        gpu_memory_pool = cp.get_default_memory_pool()
        del self.volume, self.volume_rft, self.mask, self.mask_texture, self.mask_padded, self.template, (
            self.template_texture), self.template_padded, self.wedge, self.ccc_map, self.scores, self.angles, (
            self.std_volume)
        gc.collect()
        gpu_memory_pool.free_all_blocks()


class TemplateMatchingGPU:
    def __init__(
            self,
            job_id: str,
            device_id: int,
            volume: npt.NDArray[float],
            template: npt.NDArray[float],
            mask: npt.NDArray[float],
            angle_list: list[tuple[float, float, float]],
            angle_ids: list[int],
            mask_is_spherical: bool = True,
            wedge: Optional[npt.NDArray[float]] = None,
            stats_roi: Optional[tuple[slice, slice, slice]] = None
    ):
        """Initialize a template matching run.

        Parameters
        ----------
        job_id: str
            string for job identification
        device_id: int
            GPU device id to run the job on
        volume: npt.NDArray[float]
            3D numpy array of tomogram
        template: npt.NDArray[float]
            3D numpy array of template, square box of size sx
        mask: npt.NDArray[float]
            3D numpy array with mask, same box size as template
        angle_list: list[tuple[float, float, float]]
            list of tuples with 3 floats representing Euler angle rotations
        angle_ids: list[int]
            list of indices for angle_list to actually search, this can be a subset of the full list
        mask_is_spherical: bool, default True
            True (default) if mask is spherical, set to False for non-spherical mask which increases computation time
        wedge: Optional[npt.NDArray[float]], default None
            3D numpy array that contains the Fourier space weighting for the template, it should be in Fourier
            reduced form, with dimensions (sx, sx, sx // 2 + 1)
        stats_roi: Optional[tuple[slice, slice, slice]], default None
            region of interest to calculate statistics on the search volume, default will just take the full search
            volume
        """
        cp.cuda.Device(device_id).use()

        self.job_id = job_id
        self.device_id = device_id
        self.active = True
        self.completed = False
        self.mask_is_spherical = mask_is_spherical  # whether mask is spherical
        self.angle_list = angle_list
        self.angle_ids = angle_ids
        self.stats = {'search_space': 0, 'variance': 0., 'std': 0.}
        if stats_roi is None:
            self.stats_roi = (
                slice(None),
                slice(None),
                slice(None)
            )
        else:
            self.stats_roi = stats_roi

        self.plan = TemplateMatchingPlan(volume, template, mask, device_id, wedge=wedge)

    def run(self) -> tuple[npt.NDArray[float], npt.NDArray[float], dict]:
        """Run the template matching job.

        Returns
        -------
        results: tuple[npt.NDArray[float], npt.NDArray[float], dict]
            results is a tuple with tree elements:
                - score_map with the locally normalised maximum score over all the angles searched; a 3D numpy array
                with same size as search volume
                - angle_map with an index into the angle list corresponding to the maximum of the correlation scores;
                a 3D numpy array with same size as search volume
                - a dictionary with three floats of search statistics - 'search_space', 'variance', and 'std'
        """
        print("Progress job_{} on device {:d}:".format(self.job_id, self.device_id))

        # Size x template (sxz) and center x template (cxt)
        sxt, syt, szt = self.plan.template.shape
        cxt, cyt, czt = sxt // 2, syt // 2, szt // 2
        mx, my, mz = sxt % 2, syt % 2, szt % 2  # odd or even

        # Size x volume (sxv) and center x volume (xcv)
        sxv, syv, szv = self.plan.template_padded.shape
        cxv, cyv, czv = sxv // 2, syv // 2, szv // 2

        # calculate roi size
        roi_size = self.plan.volume[self.stats_roi].size

        if self.mask_is_spherical:  # Then we only need to calculate std volume once
            self.plan.mask_padded[cxv - cxt:cxv + cxt + mx,
                                  cyv - cyt:cyv + cyt + my,
                                  czv - czt:czv + czt + mz] = self.plan.mask
            self.plan.std_volume = std_under_mask_convolution(
                self.plan.volume,
                self.plan.mask_padded,
                self.plan.mask_weight,
                volume_rft=self.plan.volume_rft
            )

        # Track iterations with a tqdm progress bar
        for i in tqdm(range(len(self.angle_ids))):

            # tqdm cannot loop over zipped lists, so need to do it like this
            angle_id, rotation = self.angle_ids[i], self.angle_list[i]

            if not self.mask_is_spherical:
                self.plan.mask_texture.transform(
                    rotation=(rotation[0], rotation[1], rotation[2]),
                    rotation_order='rzxz',
                    output=self.plan.mask,
                    rotation_units='rad'
                )
                self.plan.mask_padded[cxv - cxt:cxv + cxt + mx,
                                      cyv - cyt:cyv + cyt + my,
                                      czv - czt:czv + czt + mz] = self.plan.mask
                # Std volume needs to be recalculated for every rotation of the mask, expensive step
                self.plan.std_volume = std_under_mask_convolution(
                    self.plan.volume,
                    self.plan.mask_padded,
                    self.plan.mask_weight,
                    volume_rft=self.plan.volume_rft,
                )

            # Rotate template
            self.plan.template_texture.transform(
                rotation=(rotation[0], rotation[1], rotation[2]),
                rotation_order='rzxz',
                output=self.plan.template,
                rotation_units='rad'
            )

            if self.plan.wedge is not None:
                # Add wedge to the template after rotating
                self.plan.template = irfftn(
                    rfftn(self.plan.template) * self.plan.wedge,
                    s=self.plan.template.shape
                ).real

            # Normalize and mask template
            mean = mean_under_mask(self.plan.template, self.plan.mask, mask_weight=self.plan.mask_weight)
            std = std_under_mask(self.plan.template, self.plan.mask, mean, mask_weight=self.plan.mask_weight)
            self.plan.template = ((self.plan.template - mean) / std) * self.plan.mask

            # Paste in center
            self.plan.template_padded[cxv - cxt:cxv + cxt + mx,
                                      cyv - cyt:cyv + cyt + my,
                                      czv - czt:czv + czt + mz] = self.plan.template

            # Fast local correlation function between volume and template, norm is the standard deviation at each
            # point in the volume in the masked area
            self.plan.ccc_map = fftshift(
                irfftn(self.plan.volume_rft * rfftn(self.plan.template_padded).conj(),
                       s=self.plan.template_padded.shape).real
                / (self.plan.mask_weight * self.plan.std_volume)
            )

            # Update the scores and angle_lists
            update_results_kernel(
                self.plan.scores,
                self.plan.ccc_map,
                angle_id,
                self.plan.scores,
                self.plan.angles
            )

            self.stats['variance'] += (
                square_sum_kernel(
                    self.plan.ccc_map[self.stats_roi]
                ) / roi_size
            )

        self.stats['search_space'] = int(roi_size * len(self.angle_ids))
        self.stats['variance'] = float(self.stats['variance'] / len(self.angle_ids))
        self.stats['std'] = float(cp.sqrt(self.stats['variance']))

        # package results back to the CPU
        results = (self.plan.scores.get(), self.plan.angles.get(), self.stats)

        # clear all the used gpu memory
        self.plan.clean()

        return results


def std_under_mask_convolution(
        volume: cpt.NDArray[float],
        padded_mask: cpt.NDArray[float],
        mask_weight: float,
        volume_rft: Optional[cpt.NDArray[complex]] = None
) -> cpt.NDArray[float]:
    """Calculate the local standard deviation under the mask for each position in the volume. Calculation is done in
    Fourier space as this is a convolution between volume and mask.

    Parameters
    ----------
    volume: cpt.NDArray[float]
        cupy array to calculate local std in
    padded_mask: cpt.NDArray[float]
        template mask that has been padded to dimensions of volume
    mask_weight: float
        weight of the mask, usually calculated as mask.sum()
    volume_rft: Optional[cpt.NDArray[float]], default None
        optionally provide a precalculated reduced Fourier transform of volume to save computation

    Returns
    -------
    std_v: cpt.NDArray[float]
        array with local standard deviations in volume
    """
    volume_rft = rfftn(volume) if volume_rft is None else volume_rft
    std_v = (
            mean_under_mask_convolution(rfftn(volume ** 2), padded_mask, mask_weight) -
            mean_under_mask_convolution(volume_rft, padded_mask, mask_weight) ** 2
    )
    std_v[std_v <= cp.float32(1e-18)] = 1  # prevent potential sqrt of negative value and division by zero
    std_v = cp.sqrt(std_v)
    return std_v


def mean_under_mask_convolution(
        volume_rft: cpt.NDArray[complex],
        mask: cpt.NDArray[float],
        mask_weight: float
) -> cpt.NDArray[float]:
    """Calculate local mean in volume under the masked region.

    Parameters
    ----------
    volume_rft: cpt.NDArray[complex]
        array containing the rfftn of the volume
    mask: cpt.NDArray[float]
        mask to calculate the mean under
    mask_weight: float
        weight of the mask, usually calculated as mask.sum()

    Returns
    -------
    mean: cpt.NDArray[float]
        array with local means under the mask
    """
    return irfftn(
        volume_rft * rfftn(mask).conj(), s=mask.shape
    ).real / mask_weight


"""Update scores and angles if a new maximum is found."""
update_results_kernel = cp.ElementwiseKernel(
    'float32 scores, float32 ccc_map, float32 angle_id',
    'float32 out1, float32 out2',
    'if (scores < ccc_map) {out1 = ccc_map; out2 = angle_id;}',
    'update_results'
)


# Temporary workaround for ReductionKernel issue in cupy 13.0.0 (see: https://github.com/cupy/cupy/issues/8184)
if version.parse(cp.__version__) == version.parse('13.0.0'):
    def square_sum_kernel(x):
        return (x ** 2).sum()
else:
    """Calculate the sum of squares in a volume. Mean is assumed to be 0 which makes this operation a lot faster."""
    square_sum_kernel = cp.ReductionKernel(
        'T x',  # input params
        'T y',  # output params
        'x * x',  # pre-processing expression
        'a + b',  # reduction operation
        'y = a',  # post-reduction output processing
        '0',  # identity value
        'variance'  # kernel name
    )
